from typing import List


class BaseMetric:
    metric_names: List[str] = ['']
    primary_metric: str = ''

    def __init__(self, **kwargs):
        assert self.primary_metric in self.metric_names
        pass

    def calibrate(self, pred, target, **kwargs):
        pass



from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Union

from omegaconf import MISSING, MissingMandatoryValue

DEFAULT_IMG_SIZE = 256


@dataclass
class Transform:
    name: str = MISSING


@dataclass
class Train:
    transforms: Optional[List] = None
    mix_transforms: Optional[List] = None


@dataclass
class Inference:
    transforms: Optional[List] = None


@dataclass
class AugmentationConfig:
    img_size: int = DEFAULT_IMG_SIZE
    train: Train = field(default_factory=lambda: Train())
    inference: Inference = field(default_factory=lambda: Inference())


@dataclass
class ColorJitter(Transform):
    name: str = 'colorjitter'
    brightness: Optional[float] = 0.25
    contrast: Optional[float] = 0.25
    saturation: Optional[float] = 0.25
    hue: Optional[float] = 0.1
    p: Optional[float] = 0.5


@dataclass
class Pad(Transform):
    name: str = 'pad'
    padding: int = 0
    fill: int = 0
    padding_mode: str = 'constant'


@dataclass
class RandomCrop(Transform):
    name: str = 'randomcrop'
    size: int = DEFAULT_IMG_SIZE


@dataclass
class RandomResizedCrop(Transform):
    name: str = 'randomresizedcrop'
    size: int = DEFAULT_IMG_SIZE
    scale: List = field(default_factory=lambda: [0.08, 1.0])
    ratio: List = field(default_factory=lambda: [0.75, 1.33])
    interpolation: Optional[str] = 'bilinear'


@dataclass
class RandomHorizontalFlip(Transform):
    name: str = 'randomhorizontalflip'
    p: float = 0.5


@dataclass
class RandomVerticalFlip(Transform):
    name: str = 'randomverticalflip'
    p: float = 0.5


@dataclass
class Resize(Transform):
    name: str = 'resize'
    size: List = field(default_factory=lambda: [DEFAULT_IMG_SIZE, DEFAULT_IMG_SIZE])
    interpolation: Optional[str] = 'bilinear'
    max_size: Optional[int] =  None


class TrivialAugmentWide(Transform):
    name: str = 'trivialaugmentwide'
    num_magnitude_bins: int = 31
    interpolation: str = 'bilinear'
    fill: Optional[int] = None


@dataclass
class RandomMixup(Transform):
    name: str = 'mixup'
    alpha: float = 0.2
    p: float = 1.0
    inplace: bool = False


@dataclass
class RandomCutmix(Transform):
    name: str = 'cutmix'
    alpha: float = 1.0
    p: float = 1.0
    inplace: bool = False


@dataclass
class ClassificationAugmentationConfig(AugmentationConfig):
    img_size: int = 256
    train: Train = field(default_factory=lambda: Train(
        transforms=[RandomResizedCrop(size=256), RandomHorizontalFlip()],
        mix_transforms=[RandomCutmix()]
    ))
    inference: Inference = field(default_factory=lambda: Inference(
        transforms=[Resize(size=[256, 256])]
    ))


@dataclass
class SegmentationAugmentationConfig(AugmentationConfig):
    img_size: int = 512
    train: Train = field(default_factory=lambda: Train(
        transforms=[RandomResizedCrop(size=512), RandomHorizontalFlip(), ColorJitter()],
        mix_transforms=None
    ))
    inference: Inference = field(default_factory=lambda: Inference(
        transforms=[Resize(size=[512, 512])]
    ))


@dataclass
class DetectionAugmentationConfig(AugmentationConfig):
    img_size: int = 512
    train: Train = field(default_factory=lambda: Train(
        transforms=[Resize(size=[512, 512])],
        mix_transforms=None
    ))
    inference: Inference = field(default_factory=lambda: Inference(
        transforms=[Resize(size=[512, 512])],
    ))



from functools import partial

import cv2
import numpy as np
import PIL.Image as Image

from ..utils.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from . import custom as TC
from .registry import TRANSFORM_DICT

EDGE_SIZE = 4
Y_K_SIZE = 6
X_K_SIZE = 6


def reduce_label(label: np.ndarray) -> Image.Image:
    label[label == 0] = 255
    label = label - 1
    label[label == 254] = 255
    return Image.fromarray(label)


def generate_edge(label: np.ndarray) -> Image.Image:
    edge = cv2.Canny(label, 0.1, 0.2)
    kernel = np.ones((EDGE_SIZE, EDGE_SIZE), np.uint8)
    # edge_pad == True
    edge = edge[Y_K_SIZE:-Y_K_SIZE, X_K_SIZE:-X_K_SIZE]
    edge = np.pad(edge, ((Y_K_SIZE, Y_K_SIZE), (X_K_SIZE, X_K_SIZE)), mode='constant')
    edge = (cv2.dilate(edge, kernel, iterations=1) > 50) * 1.0
    return Image.fromarray((edge.copy() * 255).astype(np.uint8))


def transforms_custom(conf_augmentation, training):
    assert conf_augmentation.img_size > 32
    phase_conf = conf_augmentation.train if training else conf_augmentation.inference

    preprocess = []
    if phase_conf.transforms:
        for augment in phase_conf.transforms:
            name = augment.name.lower()
            augment_kwargs = list(augment.keys())
            augment_kwargs.remove('name')
            augment_kwargs = {k:augment[k] for k in augment_kwargs}
            transform = TRANSFORM_DICT[name](**augment_kwargs)
            preprocess.append(transform)

    preprocess = preprocess + [
        TC.ToTensor(),
        TC.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
    ]
    return TC.Compose(preprocess)


def train_transforms_pidnet(conf_augmentation, training):
    preprocess = []
    phase_conf = conf_augmentation.train if training else conf_augmentation.inference

    if phase_conf.transforms:
        for augment in phase_conf.transforms:
            name = augment.name.lower()
            augment_kwargs = list(augment.keys())
            augment_kwargs.remove('name')
            augment_kwargs = {k:augment[k] for k in augment_kwargs}
            transform = TRANSFORM_DICT[name](**augment_kwargs)
            preprocess.append(transform)

    preprocess = preprocess + [
        TC.ToTensor(),
        TC.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
    ]
    return TC.Compose(preprocess, additional_targets={'edge': 'mask'})


def create_transform(model_name: str, is_training=False):
    if 'pidnet' in model_name:
        return partial(train_transforms_pidnet, training=is_training)
    return partial(transforms_custom, training=is_training)



import os
from pathlib import Path
from typing import Callable, Dict, List, Type

import torch
import torch.nn as nn
from loguru import logger
from omegaconf import OmegaConf

from .base import ClassificationModel, DetectionModel, SegmentationModel, TaskModel
from .registry import (
    MODEL_BACKBONE_DICT,
    MODEL_FULL_DICT,
    MODEL_HEAD_DICT,
    MODEL_NECK_DICT,
    SUPPORTING_TASK_LIST,
    TASK_MODEL_DICT,
)
from .utils import load_from_checkpoint


def load_full_model(conf_model, model_name, num_classes, model_checkpoint, use_pretrained):
    model_fn: Callable[..., nn.Module] = MODEL_FULL_DICT[model_name]
    conf_model.architecture.full.nick_name = conf_model.name # @illian01: model.name should be same with conf_model.name
    model: nn.Module = model_fn(num_classes=num_classes, conf_model_full=conf_model.architecture.full)
    if use_pretrained:
        model = load_from_checkpoint(
            model, model_checkpoint,
            load_checkpoint_head=conf_model.checkpoint.load_head,
        )

    return model


def load_backbone_and_head_model(
        conf_model, task, backbone_name, head_name, num_classes,
        model_checkpoint, use_pretrained, freeze_backbone,
        img_size
    ):
    if task not in TASK_MODEL_DICT:
        raise ValueError(
            f"No such task(s) named: {task}. This should be included in SUPPORTING_TASK_LIST ({SUPPORTING_TASK_LIST})")

    # Backbone construction
    backbone_fn: Callable[..., nn.Module] = MODEL_BACKBONE_DICT[backbone_name]
    backbone: nn.Module = backbone_fn(task=task, conf_model_backbone=conf_model.architecture.backbone)

    # Neck construction
    intermediate_features_dim = backbone.intermediate_features_dim
    neck = None
    if getattr(conf_model.architecture, 'neck', None):
        neck_name = conf_model.architecture.neck.name
        neck_fn: Callable[..., nn.Module] = MODEL_NECK_DICT[neck_name]
        neck = neck_fn(intermediate_features_dim=backbone.intermediate_features_dim, conf_model_neck=conf_model.architecture.neck)
        intermediate_features_dim = neck.intermediate_features_dim

    # Head construction
    head_module = MODEL_HEAD_DICT[task][head_name]
    if task == 'classification':
        head = head_module(num_classes=num_classes, feature_dim=backbone.feature_dim, conf_model_head=conf_model.architecture.head)
    elif task in ['segmentation', 'detection']:
        img_size = img_size if isinstance(img_size, (int, None)) else tuple(img_size)
        head = head_module(num_classes=num_classes,
                                intermediate_features_dim=intermediate_features_dim,
                                label_size=img_size,
                                conf_model_head=conf_model.architecture.head)

    # Assemble model and load checkpoint
    model = TASK_MODEL_DICT[task](conf_model, backbone, neck, head, freeze_backbone)
    if use_pretrained:
        model = load_from_checkpoint(
            model, model_checkpoint,
            load_checkpoint_head=conf_model.checkpoint.load_head,
        )
    return model


def build_model(conf_model, task, num_classes, model_checkpoint, use_pretrained, img_size) -> nn.Module:

    if conf_model.single_task_model:
        model_name = str(conf_model.architecture.full.name).lower()
        return load_full_model(
            conf_model, model_name, num_classes,
            model_checkpoint=model_checkpoint,
            use_pretrained=use_pretrained
        )

    backbone_name = str(conf_model.architecture.backbone.name).lower()
    head_name = str(conf_model.architecture.head.name).lower()
    freeze_backbone = conf_model.freeze_backbone
    return load_backbone_and_head_model(
        conf_model, task, backbone_name, head_name, num_classes,
        model_checkpoint=model_checkpoint,
        use_pretrained=use_pretrained,
        freeze_backbone=freeze_backbone,
        img_size=img_size
    )



from typing import Callable, Dict, Type

from .augmentation.transforms import create_transform
from .base import BaseCustomDataset, BaseDataSampler, BaseHFDataset
from .classification import (
    ClassficationDataSampler,
    ClassificationCustomDataset,
    ClassificationHFDataset,
)
from .detection import DetectionCustomDataset, DetectionDataSampler
from .segmentation import (
    SegmentationCustomDataset,
    SegmentationDataSampler,
    SegmentationHFDataset,
)

CREATE_TRANSFORM = create_transform

CUSTOM_DATASET: Dict[str, Type[BaseCustomDataset]] = {
    'classification': ClassificationCustomDataset,
    'segmentation': SegmentationCustomDataset,
    'detection': DetectionCustomDataset
}

HUGGINGFACE_DATASET: Dict[str, Type[BaseHFDataset]] = {
    'classification': ClassificationHFDataset,
    'segmentation': SegmentationHFDataset
}

DATA_SAMPLER: Dict[str, Type[BaseDataSampler]] = {
    'classification': ClassficationDataSampler,
    'segmentation': SegmentationDataSampler,
    'detection': DetectionDataSampler
}



import os
from abc import ABC, abstractmethod, abstractproperty
from itertools import repeat
from pathlib import Path

import numpy as np
import torch
import torch.utils.data as data


class BaseCustomDataset(data.Dataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class, split, samples, transform, with_label, **kwargs):
        super(BaseCustomDataset, self).__init__()
        self.conf_data = conf_data
        self.conf_augmentation = conf_augmentation
        self.model_name = model_name

        self.transform = transform
        self.samples = samples

        self._root = conf_data.path.root
        self._idx_to_class = idx_to_class
        self._num_classes = len(self._idx_to_class)
        self._split = split
        self._with_label = with_label

    @abstractmethod
    def __getitem__(self, index):
        pass

    def __len__(self):
        return len(self.samples)

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def class_map(self):
        return self._idx_to_class

    @property
    def root(self):
        return self._root

    @property
    def mode(self):
        return self._split

    @property
    def with_label(self):
        return self._with_label


class BaseHFDataset(data.Dataset):

    def __init__(self, conf_data, conf_augmentation, model_name, root, split, with_label):
        super(BaseHFDataset, self).__init__()
        self.conf_data = conf_data
        self.conf_augmentation = conf_augmentation
        self.model_name = model_name
        self._root = root
        self._split = split
        self._with_label = with_label

    def _load_dataset(self, root, subset_name=None, cache_dir=None):
        from datasets import load_dataset
        if cache_dir is not None:
            Path(cache_dir).mkdir(exist_ok=True, parents=True)
        total_dataset = load_dataset(root, name=subset_name, cache_dir=cache_dir)
        return total_dataset

    @abstractmethod
    def __getitem__(self, index):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractproperty
    def num_classes(self):
        pass

    @abstractproperty
    def class_map(self):
        pass

    @property
    def root(self):
        return self._name

    @property
    def mode(self):
        return self._split

    @property
    def with_label(self):
        return self._with_label


class BaseDataSampler(ABC):
    def __init__(self, conf_data, train_valid_split_ratio):
        self.conf_data = conf_data
        self.train_valid_split_ratio = train_valid_split_ratio

    @abstractmethod
    def load_data(self):
        raise NotImplementedError

    @abstractmethod
    def load_samples(self):
        raise NotImplementedError

    @abstractmethod
    def load_huggingface_samples(self):
        raise NotImplementedError



from typing import Dict, Type

import torch
import torch.nn as nn

NORM_REGISTRY: Dict[str, Type[nn.Module]] = {
    'batch_norm': nn.BatchNorm2d,
    'instance_norm': nn.InstanceNorm2d,
}

ACTIVATION_REGISTRY: Dict[str, Type[nn.Module]] = {
    'relu': nn.ReLU,
    'prelu': nn.PReLU,
    'leaky_relu': nn.LeakyReLU,
    'gelu': nn.GELU,
    'silu': nn.SiLU,
    'swish': nn.SiLU,
    'hard_swish': nn.Hardswish,
}



import os

import requests
from bson.objectid import ObjectId
from loguru import logger
from pymongo.mongo_client import MongoClient

MONGODB_TEMP_URI = ""


class ModelSearchServerHandler:
    def __init__(self, task, model, mongodb_uri: str=MONGODB_TEMP_URI) -> None:
        client = MongoClient(mongodb_uri)

        try:
            client.admin.command('ping')
            logger.debug("Pinged your deployment. You successfully connected to MongoDB!")
        except Exception as e:
            raise e

        self._db = client['custom-training-board']['trainer-all-in-one']
        self._session_id = None

        self._create_session(title=f"[{task}]{model}")


    def init_epoch(self):
        self._epoch = 0

    @property
    def epoch(self):
        return self._epoch

    @epoch.setter
    def epoch(self, value: int) -> None:
        self._epoch = int(value)

    def _is_ready(self):
        return self._session_id is not None

    def _append(self, scalar_dict, mode='train'):
        assert self._is_ready()
        meta_string = f"{mode}/" if mode is not None else ""
        contents = {'$push': {f"{meta_string}{key}": {'epoch': self._epoch, 'value': value}
                              for key, value in scalar_dict.items()},
                    '$currentDate': {'lastModified': True }}
        result = self._db.update_one({'_id': self._session_id}, contents, upsert=True)
        return result

    def _create_session(self, title: str ="test") -> ObjectId:
        example_document = { "title": title }
        document = self._db.insert_one(example_document)
        self._session_id = document.inserted_id
        return self._session_id

    def create_session(self, title: str="test") -> ObjectId:
        return self._create_session(title=title)

    def log_scalar(self, key, value, mode='train'):
        result = self._append({key: value}, mode=mode)
        return result

    def log_scalars_with_dict(self, scalar_dict, mode='train'):
        result = self._append(scalar_dict, mode=mode)
        return result

    def __call__(self,
            train_losses, train_metrics, valid_losses, valid_metrics,
            learning_rate, elapsed_time,
        ) -> None:

        self.log_scalars_with_dict(train_losses, mode='train')
        self.log_scalars_with_dict(train_metrics, mode='train')

        if valid_losses is not None:
            self.log_scalars_with_dict(valid_losses, mode='valid')
        if valid_metrics is not None:
            self.log_scalars_with_dict(valid_metrics, mode='valid')

        if learning_rate is not None:
            self.log_scalar('learning_rate', learning_rate, mode='misc')
        if elapsed_time is not None:
            self.log_scalar('elapsed_time', elapsed_time, mode='misc')



import logging
from typing import TypedDict

import thop
import torch
import torch.nn as nn
from fvcore.nn import FlopCountAnalysis

from .environment import get_device


def get_params_and_macs(model: nn.Module, sample_input: torch.Tensor):
    sample_input = sample_input.to(get_device(model))
    # From v0.0.9
    macs, params = _params_and_macs_fvcore(model, sample_input)

    # # Before v0.0.9
    # macs, params = _params_and_macs_thop(model, sample_input)

    return macs, params

def _params_and_macs_fvcore(model: nn.Module, sample_input: torch.Tensor):
    fvcore_logger = logging.getLogger('fvcore')
    fvcore_logger.setLevel(logging.CRITICAL)
    macs = FlopCountAnalysis(model, sample_input).total()
    params = sum(p.numel() for p in model.parameters())
    return macs, params

def _params_and_macs_thop(model: nn.Module, sample_input: torch.Tensor):
    macs, params = thop.profile(model, inputs=(sample_input,), verbose=False)
    return macs, params



import math
from typing import List, Tuple, Union

from omegaconf import DictConfig
import torch
import torch.nn as nn
import torch.nn.functional as F

from ....op.custom import ConvLayer
from ....utils import FXTensorListType, ModelOutput


class AllMLPDecoder(nn.Module):
    def __init__(
            self,
            num_classes: int,
            intermediate_features_dim: List[int],
            label_size: Union[Tuple[int, int], int],
            params: DictConfig,
        ):
        super().__init__()
        intermediate_channels = params.intermediate_channels
        classifier_dropout_prob = params.classifier_dropout_prob

        # linear layers which will unify the channel dimension of each of the encoder blocks to the same config.decoder_hidden_size
        mlps = []
        for feature_dim in intermediate_features_dim:
            mlp = ConvLayer(feature_dim, intermediate_channels, kernel_size=1,
                            use_norm=True, use_act=True, norm_type='batch_norm', act_type='relu')
            mlps.append(mlp)
        self.linear_c = nn.ModuleList(mlps)

        # the following 3 layers implement the ConvModule of the original implementation
        self.linear_fuse = ConvLayer(len(intermediate_features_dim) * intermediate_channels,
                                     intermediate_channels, kernel_size=1,
                                     use_norm=True, use_act=False)

        self.dropout = nn.Dropout(classifier_dropout_prob)
        self.classifier = nn.Conv2d(intermediate_channels, num_classes, kernel_size=1)

        self.label_size = (label_size, label_size) if isinstance(label_size, int) else label_size

    def forward(self, encoder_hidden_states: FXTensorListType):

        all_hidden_states = ()
        for encoder_hidden_state, mlp in zip(encoder_hidden_states, self.linear_c):

            # unify channel dimension
            encoder_hidden_state = mlp(encoder_hidden_state)

            encoder_hidden_state = F.interpolate(
                encoder_hidden_state, size=encoder_hidden_states[0].size()[2:], mode="bilinear", align_corners=False
            )  # upsample to H/4 x W/4
            all_hidden_states += (encoder_hidden_state,)

        hidden_states = self.linear_fuse(torch.cat(all_hidden_states[::-1], dim=1))
        hidden_states = self.dropout(hidden_states)

        # logits are of shape (batch_size, num_labels, height/4, width/4)
        logits = self.classifier(hidden_states)  # B x {num_classes} x H/4 x W/4

        if self.label_size is not None:
            H, W = self.label_size[-2:]
            # upsample logits to the images' original size
            logits = F.interpolate(
                logits, size=(H, W), mode="bilinear", align_corners=False
            )

        return ModelOutput(pred=logits)


def all_mlp_decoder(num_classes, intermediate_features_dim, label_size, conf_model_head, **kwargs) -> AllMLPDecoder:
    return AllMLPDecoder(num_classes,
                         intermediate_features_dim,
                         label_size=label_size,
                         params=conf_model_head.params)



# from .core import *
from .core.resnet import resnet
from .experimental.darknet import cspdarknet
from .experimental.efficientformer import efficientformer
from .experimental.mixnet import mixnet
from .experimental.mixtransformer import mixtransformer
from .experimental.mobilenetv3 import mobilenetv3
from .experimental.mobilevit import mobilevit
from .experimental.vit import vit



import os
from pathlib import Path
from typing import Literal

import numpy as np
import PIL.Image as Image

from ..augmentation.transforms import generate_edge, reduce_label
from ..base import BaseCustomDataset


class SegmentationCustomDataset(BaseCustomDataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class,
                 split, samples, transform=None, with_label=True, **kwargs):
        super(SegmentationCustomDataset, self).__init__(
            conf_data, conf_augmentation, model_name, idx_to_class,
            split, samples, transform, with_label, **kwargs
        )
        assert "label_value_to_idx" in kwargs
        self.label_value_to_idx = kwargs["label_value_to_idx"]

        self.label_image_mode: Literal['RGB', 'L', 'P'] = str(conf_data.label_image_mode).upper() \
            if conf_data.label_image_mode is not None else 'L'

    def __getitem__(self, index):
        img_path = Path(self.samples[index]['image'])
        ann_path = Path(self.samples[index]['label']) if 'label' in self.samples[index] else None
        img = Image.open(img_path).convert('RGB')

        org_img = img.copy()

        w, h = img.size

        if ann_path is None:
            out = self.transform(self.conf_augmentation)(image=img)
            return {'pixel_values': out['image'], 'name': img_path.name, 'org_img': org_img, 'org_shape': (h, w)}

        outputs = {}

        label = Image.open(ann_path).convert(self.label_image_mode)
        label_array = np.array(label)
        label_array = label_array[..., np.newaxis] if label_array.ndim == 2 else label_array
        # if self.conf_augmentation.reduce_zero_label:
        #     label = reduce_label(np.array(label))

        mask = np.zeros((label.size[1], label.size[0]), dtype=np.uint8)
        for label_value in self.label_value_to_idx:
            class_mask = (label_array == np.array(label_value)).all(axis=-1)
            mask[class_mask] = self.label_value_to_idx[label_value]

        mask = Image.fromarray(mask, mode='L')  # single mode array (PIL.Image) compatbile with torchvision transform API

        if 'pidnet' in self.model_name:
            edge = generate_edge(np.array(mask))
            out = self.transform(self.conf_augmentation)(image=img, mask=mask, edge=edge)
            outputs.update({'pixel_values': out['image'], 'labels': out['mask'], 'edges': out['edge'].float(), 'name': img_path.name})
        else:
            out = self.transform(self.conf_augmentation)(image=img, mask=mask)
            outputs.update({'pixel_values': out['image'], 'labels': out['mask'], 'name': img_path.name})

        if self._split in ['train', 'training']:
            return outputs

        assert self._split in ['val', 'valid', 'test']
        # outputs.update({'org_img': org_img, 'org_shape': (h, w)})  # TODO: return org_img with batch_size > 1
        outputs.update({'org_shape': (h, w)})
        return outputs



"""
Based on the Darknet implementation of Megvii.
https://github.com/Megvii-BaseDetection/YOLOX/blob/main/yolox/models/darknet.py
"""
from typing import Dict, Optional, List

from omegaconf import DictConfig
import torch
from torch import nn

from ...op.custom import ConvLayer, CSPLayer, Focus, SPPBottleneck
from ...utils import BackboneOutput

__all__ = ['cspdarknet']
SUPPORTING_TASK = ['detection']


class CSPDarknet(nn.Module):

    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
        #depthwise=False,
    ) -> None:
        super().__init__()
        out_features=("dark3", "dark4", "dark5")
        assert out_features, "please provide output features of Darknet"

        self.task = task.lower()
        self.use_intermediate_features = self.task in ['segmentation', 'detection']

        dep_mul = params.dep_mul
        wid_mul = params.wid_mul
        act_type = params.act_type

        self.out_features = out_features
        Conv = ConvLayer

        base_channels = int(wid_mul * 64)  # 64
        base_depth = max(round(dep_mul * 3), 1)  # 3

        # stem
        self.stem = Focus(3, base_channels, ksize=3, act_type=act_type)

        # dark2
        self.dark2 = nn.Sequential(
            Conv(in_channels=base_channels, 
                 out_channels=base_channels * 2, 
                 kernel_size=3, 
                 stride=2, 
                 act_type=act_type),
            CSPLayer(
                base_channels * 2,
                base_channels * 2,
                n=base_depth,
                #depthwise=depthwise,
                act_type=act_type,
            ),
        )

        # dark3
        self.dark3 = nn.Sequential(
            Conv(in_channels=base_channels * 2, 
                 out_channels=base_channels * 4, 
                 kernel_size=3, 
                 stride=2, 
                 act_type=act_type),
            CSPLayer(
                base_channels * 4,
                base_channels * 4,
                n=base_depth * 3,
                #depthwise=depthwise,
                act_type=act_type,
            ),
        )

        # dark4
        self.dark4 = nn.Sequential(
            Conv(in_channels=base_channels * 4,
                 out_channels=base_channels * 8, 
                 kernel_size=3, 
                 stride=2, 
                 act_type=act_type),
            CSPLayer(
                base_channels * 8,
                base_channels * 8,
                n=base_depth * 3,
                #depthwise=depthwise,
                act_type=act_type,
            ),
        )

        # dark5
        self.dark5 = nn.Sequential(
            Conv(in_channels=base_channels * 8, 
                 out_channels=base_channels * 16, 
                 kernel_size=3, 
                 stride=2, 
                 act_type=act_type),
            SPPBottleneck(base_channels * 16, base_channels * 16, act_type=act_type),
            CSPLayer(
                base_channels * 16,
                base_channels * 16,
                n=base_depth,
                shortcut=False,
                #depthwise=depthwise,
                act_type=act_type,
            ),
        )

        self.avgpool = nn.AdaptiveAvgPool2d(1)

        predefined_out_features = {'dark2': base_channels * 2, 'dark3': base_channels * 4, 
                                   'dark4': base_channels * 8, 'dark5': base_channels * 16}
        self._feature_dim = predefined_out_features['dark5']
        self._intermediate_features_dim = [predefined_out_features[out_feature] for out_feature in out_features]

    def forward(self, x):
        outputs_dict = {}
        x = self.stem(x)
        outputs_dict["stem"] = x
        x = self.dark2(x)
        outputs_dict["dark2"] = x
        x = self.dark3(x)
        outputs_dict["dark3"] = x
        x = self.dark4(x)
        outputs_dict["dark4"] = x
        x = self.dark5(x)
        outputs_dict["dark5"] = x

        if self.use_intermediate_features:
            all_hidden_states = [outputs_dict[out_name] for out_name in self.out_features]
            return BackboneOutput(intermediate_features=all_hidden_states)
        
        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        return BackboneOutput(last_feature=x)
    
    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def cspdarknet(task, conf_model_backbone) -> CSPDarknet:
    return CSPDarknet(task, conf_model_backbone.params, conf_model_backbone.stage_params)



import os
from pathlib import Path
from typing import List

import numpy as np
import PIL.Image as Image
import torch
from omegaconf import OmegaConf

from ..base import BaseCustomDataset

ID2LABEL_FILENAME = "id2label.json"
TEMP_COCO_LABEL_FILE = "data/detection/coco.yaml"


def exist_name(candidate, folder_iterable):
    try:
        return list(filter(lambda x: candidate[0] in x, folder_iterable))[0]
    except IndexError:
        return list(filter(lambda x: candidate[1] in x, folder_iterable))[0]


def get_label(label_file: Path):
    target = Path(label_file).read_text()

    try:
        target_array = np.array([list(map(float, box.split(' '))) for box in target.split('\n') if box.strip()])
    except ValueError as e:
        print(target)
        raise e

    label, boxes = target_array[:, 0], target_array[:, 1:]
    label = label[..., np.newaxis]
    return label, boxes


class DetectionCustomDataset(BaseCustomDataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class,
                 split, samples, transform=None, with_label=True, **kwargs):
        super(DetectionCustomDataset, self).__init__(
            conf_data, conf_augmentation, model_name, idx_to_class,
            split, samples, transform, with_label, **kwargs
        )

    @staticmethod
    def xywhn2xyxy(original: np.ndarray, w: int, h: int, padw=0, padh=0):
        converted = original.copy()
        # left, top (lt)
        converted[..., 0] = w * (original[..., 0] - original[..., 2] / 2) + padw
        converted[..., 1] = h * (original[..., 1] - original[..., 3] / 2) + padh
        # right, bottom (rb)
        converted[..., 2] = w * (original[..., 0] + original[..., 2] / 2) + padw
        converted[..., 3] = h * (original[..., 1] + original[..., 3] / 2) + padh
        return converted

    def __getitem__(self, index):
        img_path = Path(self.samples[index]['image'])
        ann_path = Path(self.samples[index]['label']) if 'label' in self.samples[index] else None
        img = Image.open(str(img_path)).convert('RGB')

        org_img = img.copy()

        w, h = img.size

        if ann_path is None:
            out = self.transform(self.conf_augmentation)(image=img)
            return {'pixel_values': out['image'], 'name': img_path.name, 'org_img': org_img, 'org_shape': (h, w)}

        outputs = {}

        label, boxes_yolo = get_label(Path(ann_path))
        boxes = self.xywhn2xyxy(boxes_yolo, w, h)

        out = self.transform(self.conf_augmentation)(image=img, bbox=np.concatenate((boxes, label), axis=-1))
        assert out['bbox'].shape[-1] == 5  # ltrb + class_label
        outputs.update({'pixel_values': out['image'], 'bbox': out['bbox'][..., :4],
                        'label': torch.as_tensor(out['bbox'][..., 4], dtype=torch.int64)})


        if self._split in ['train', 'training']:
            return outputs

        assert self._split in ['val', 'valid', 'test']
        # outputs.update({'org_img': org_img, 'org_shape': (h, w)})  # TODO: return org_img with batch_size > 1
        outputs.update({'org_shape': (h, w)})
        return outputs



from .common import CrossEntropyLoss, SigmoidFocalLoss
from .detection import RetinaNetLoss, YOLOXLoss
from .segmentation import PIDNetLoss

LOSS_DICT = {
    'cross_entropy': CrossEntropyLoss,
    'pidnet_loss': PIDNetLoss,
    'yolox_loss': YOLOXLoss,
    'retinanet_loss': RetinaNetLoss,
    'focal_loss': SigmoidFocalLoss,
}

PHASE_LIST = ['train', 'valid', 'test']



"""
Based on the Torchvision implementation of MobileNetV3.
https://pytorch.org/vision/main/_modules/torchvision/models/mobilenetv3.html
"""
from typing import List, Dict, Optional

from omegaconf import DictConfig
import torch
import torch.nn as nn
from torch import Tensor

from ...op.custom import ConvLayer, InvertedResidual
from ...utils import BackboneOutput

__all__ = ['mobilenetv3']

SUPPORTING_TASK = ['classification', 'segmentation']


class MobileNetV3(nn.Module):

    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        super(MobileNetV3, self).__init__()

        self.task = task.lower()
        self.use_intermediate_features = self.task in ['segmentation', 'detection']
        norm_type = 'batch_norm'
        act_type = 'hard_swish'

        # building first layer
        firstconv_output_channels = stage_params[0].in_channels[0]
        self.conv_first = ConvLayer(
            in_channels=3,
            out_channels=firstconv_output_channels,
            kernel_size=3,
            stride=2,
            norm_type=norm_type,
            act_type=act_type,
        )

        # building inverted residual blocks
        stages: List[nn.Module] = []

        lastconv_input_channels = stage_params[-1].out_channels[-1]
        lastconv_output_channels = 6 * lastconv_input_channels
        for stg_idx, stage_info in enumerate(stage_params):
            stage: List[nn.Module] = []

            for block in zip(stage_info.in_channels, stage_info.kernel_sizes, stage_info.expanded_channels,
                             stage_info.out_channels, stage_info.use_se, stage_info.act_type,
                             stage_info.stride):
                in_channels, kernel_size, hidden_channels, out_channels, use_se, act_type_b, stride = block
                dilation = 1 # Fix as a constant
                act_type_b = act_type_b.lower()
                stage.append(
                    InvertedResidual(in_channels=in_channels,
                                     hidden_channels=hidden_channels,
                                     out_channels=out_channels,
                                     kernel_size=kernel_size,
                                     stride=stride,
                                     norm_type=norm_type,
                                     act_type=act_type_b,
                                     use_se=use_se,
                                     dilation=dilation)
                )
            
            # add last conv
            if stg_idx == len(stage_params) - 1:
                stage.append(
                    ConvLayer(in_channels=lastconv_input_channels,
                              out_channels=lastconv_output_channels,
                              kernel_size=1,
                              norm_type=norm_type,
                              act_type=act_type,)
                )

            stage = nn.Sequential(*stage)
            stages.append(stage)
        
        self.stages = nn.ModuleList(stages)
        self.avgpool = nn.AdaptiveAvgPool2d(1)

        self._feature_dim = lastconv_output_channels
        self._intermediate_features_dim = [s[-1].out_channels for s in self.stages[:-1]]
        self._intermediate_features_dim += [lastconv_output_channels]

        # init weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def forward(self, x: Tensor):
        x = self.conv_first(x)

        all_hidden_states = () if self.use_intermediate_features else None
        for stage in self.stages:
            x = stage(x)
            if self.use_intermediate_features:
                all_hidden_states = all_hidden_states + (x, )
        
        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        return BackboneOutput(last_feature=x)

    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def mobilenetv3(task, conf_model_backbone) -> MobileNetV3:
    return MobileNetV3(task, conf_model_backbone.params, conf_model_backbone.stage_params)



from .csv import ClassificationCSVLogger, DetectionCSVLogger, SegmentationCSVLogger
from .visualizer import DetectionVisualizer, SegmentationVisualizer

CSV_LOGGER = {
    'classification': ClassificationCSVLogger,
    'segmentation': SegmentationCSVLogger,
    'detection': DetectionCSVLogger
}

VISUALIZER = {
    'segmentation': SegmentationVisualizer,
    'detection': DetectionVisualizer,
}



from pathlib import Path

from netspresso_trainer.cfg import TrainerConfig
from netspresso_trainer.trainer_cli import parse_args_netspresso, train_cli
from netspresso_trainer.trainer_inline import export_config_as_yaml, train_with_config, train_with_yaml
from netspresso_trainer.trainer_util import validate_config

### Starting from v0.0.9, the default train function runs with yaml configuration
train = train_with_yaml
# train = train_with_config

version = (Path(__file__).parent / "VERSION").read_text().strip()

__version__ = version



from typing import Literal

import numpy as np
import torch
from loguru import logger

from .base import BasePipeline


class DetectionPipeline(BasePipeline):
    def __init__(self, conf, task, model_name, model, devices,
                 train_dataloader, eval_dataloader, class_map, logging_dir, **kwargs):
        super(DetectionPipeline, self).__init__(conf, task, model_name, model, devices,
                                                train_dataloader, eval_dataloader, class_map, logging_dir, **kwargs)
        self.num_classes = train_dataloader.dataset.num_classes

    def train_step(self, batch):
        self.model.train()
        images, labels, bboxes = batch['pixel_values'], batch['label'], batch['bbox']
        images = images.to(self.devices)
        targets = [{"boxes": box.to(self.devices), "labels": label.to(self.devices),}
                   for box, label in zip(bboxes, labels)]

        targets = {'gt': targets,
                   'img_size': images.size(-1),
                   'num_classes': self.num_classes,}

        self.optimizer.zero_grad()

        out = self.model(images)
        self.loss_factory.calc(out, targets, phase='train')

        self.loss_factory.backward()
        self.optimizer.step()

        pred = self.postprocessor(out, original_shape=images[0].shape)

        if self.conf.distributed:
            torch.distributed.barrier()

        logs = {
            'target': [(bbox.detach().cpu().numpy(), label.detach().cpu().numpy())
                       for bbox, label in zip(bboxes, labels)],
            'pred': pred
        }
        return dict(logs.items())

    def valid_step(self, batch):
        self.model.eval()
        images, labels, bboxes = batch['pixel_values'], batch['label'], batch['bbox']
        images = images.to(self.devices)
        targets = [{"boxes": box.to(self.devices), "labels": label.to(self.devices)}
                   for box, label in zip(bboxes, labels)]

        targets = {'gt': targets,
                   'img_size': images.size(-1),
                   'num_classes': self.num_classes,}

        self.optimizer.zero_grad()

        out = self.model(images)
        self.loss_factory.calc(out, targets, phase='valid')

        pred = self.postprocessor(out, original_shape=images[0].shape)

        if self.conf.distributed:
            torch.distributed.barrier()

        logs = {
            'images': images.detach().cpu().numpy(),
            'target': [(bbox.detach().cpu().numpy(), label.detach().cpu().numpy())
                       for bbox, label in zip(bboxes, labels)],
            'pred': pred
        }
        return dict(logs.items())

    def test_step(self, batch):
        self.model.eval()
        images = batch['pixel_values']
        images = images.to(self.devices)

        out = self.model(images.unsqueeze(0))

        pred = self.postprocessor(out, original_shape=images[0].shape)

        results = pred
        return results

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid']):
        pred = []
        targets = []
        for output_batch in outputs:
            for detection, class_idx in output_batch['target']:
                target_on_image = {}
                target_on_image['boxes'] = detection
                target_on_image['labels'] = class_idx
                targets.append(target_on_image)

            for detection, class_idx in output_batch['pred']:
                pred_on_image = {}
                pred_on_image['post_boxes'] = detection[..., :4]
                pred_on_image['post_scores'] = detection[..., -1]
                pred_on_image['post_labels'] = class_idx
                pred.append(pred_on_image)
        self.metric_factory.calc(pred, target=targets, phase=phase)



"""
Based on the Torchvision implementation of ResNet.
https://pytorch.org/vision/stable/_modules/torchvision/models/resnet.html
"""
from typing import Dict, List, Literal, Optional, Type, Union

from omegaconf import DictConfig
import torch
import torch.nn as nn
from torch import Tensor

from ...op.custom import BasicBlock, Bottleneck, ConvLayer
from ...utils import BackboneOutput

__all__ = ['resnet']

SUPPORTING_TASK = ['classification', 'segmentation']

BLOCK_FROM_LITERAL: Dict[str, Type[nn.Module]] = {
    'basicblock': BasicBlock,
    'bottleneck': Bottleneck,
}


class ResNet(nn.Module):

    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        super(ResNet, self).__init__()

        block: Literal['basicblock', 'bottleneck'] = params.block_type
        norm_layer: Optional[str] = params.norm_type

        # Fix as constant
        zero_init_residual: bool = False
        groups: int = 1
        width_per_group: int = 64
        expansion: Optional[int] = None

        self.task = task.lower()
        block = BLOCK_FROM_LITERAL[block.lower()]
        self.use_intermediate_features = self.task in ['segmentation', 'detection']

        if norm_layer is None:
            norm_layer = 'batch_norm'
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        self.groups = groups
        self.base_width = width_per_group

        if expansion is None:
            expansion = block.expansion

        self.conv1 = ConvLayer(in_channels=3, out_channels=self.inplanes,
                               kernel_size=7, stride=2, padding=3,
                               bias=False, norm_type='batch_norm', act_type='relu')
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        stages: List[nn.Module] = []

        first_stage = stage_params[0]
        layer = self._make_layer(block, first_stage.channels, first_stage.num_blocks, expansion=expansion)
        stages.append(layer)
        for stage in stage_params[1:]:
            layer = self._make_layer(block, stage.channels, stage.num_blocks, stride=2,
                                     dilate=stage.replace_stride_with_dilation,
                                     expansion=expansion)
            stages.append(layer)

        self.stages = nn.ModuleList(stages)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        hidden_sizes = [stage.channels * expansion for stage in stage_params]
        self._feature_dim = hidden_sizes[-1]
        self._intermediate_features_dim = hidden_sizes

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(self, block: Type[Union[BasicBlock, Bottleneck]], planes: int, blocks: int,
                    stride: int = 1, dilate: bool = False, expansion: Optional[int] = None) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if expansion is None:
            expansion = block.expansion
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * expansion:
            downsample = ConvLayer(
                in_channels=self.inplanes, out_channels=planes * expansion,
                kernel_size=1, stride=stride, bias=False,
                norm_type=norm_layer, use_act=False
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def forward(self, x: Tensor):
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.maxpool(x)

        all_hidden_states = () if self.use_intermediate_features else None
        for stage in self.stages:
            x = stage(x)
            if self.use_intermediate_features:
                all_hidden_states = all_hidden_states + (x,)

        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        # x = self.fc(x)

        return BackboneOutput(last_feature=x)

    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def resnet(task, conf_model_backbone) -> ResNet:
    """
        ResNet model from "Deep Residual Learning for Image Recognition" https://arxiv.org/pdf/1512.03385.pdf.
    """
    return ResNet(task, conf_model_backbone.params, conf_model_backbone.stage_params)



import os

from netspresso_trainer.trainer_common import train_common
from netspresso_trainer.trainer_util import parse_args_netspresso, set_arguments

LOG_LEVEL = os.getenv('LOG_LEVEL', 'INFO')


def train_cli_without_additional_gpu_check() -> None:
    args_parsed = parse_args_netspresso(with_gpus=False)

    conf = set_arguments(
        data=args_parsed.data,
        augmentation=args_parsed.augmentation,
        model=args_parsed.model,
        training=args_parsed.training,
        logging=args_parsed.logging,
        environment=args_parsed.environment
    )

    train_common(
        conf,
        task=args_parsed.task,
        model_name=args_parsed.model_name,
        is_graphmodule_training=args_parsed.is_graphmodule_training,
        logging_dir=args_parsed.logging_dir,
        log_level=args_parsed.log_level
    )


if __name__ == "__main__":

    # Execute by `run_distributed_training_script`
    train_cli_without_additional_gpu_check()



from typing import List

import torch

from ..base import BaseMetric

TOPK_MAX = 20


@torch.no_grad()
def accuracy_topk(pred, target):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    batch_size = target.size(0)
    maxk = pred.size(-1)
    pred = pred.t()
    correct = pred.eq(target.reshape(1, -1).expand_as(pred))
    return lambda topk: correct[:min(topk, maxk)].reshape(-1).float().sum(0) * 100. / batch_size


class ClassificationMetric(BaseMetric):
    metric_names: List[str] = ['Acc@1', 'Acc@5']
    primary_metric: str = 'Acc@1'

    def __init__(self, **kwargs):
        super().__init__()

    def calibrate(self, pred, target, **kwargs):
        result_dict = {k: 0. for k in self.metric_names}
        topk_callable = accuracy_topk(pred, target)

        result_dict['Acc@1'] = topk_callable(topk=1)
        result_dict['Acc@5'] = topk_callable(topk=5)
        return result_dict



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from omegaconf import MISSING, MissingMandatoryValue

__all__ = [
    "ModelConfig",
    "ClassificationEfficientFormerModelConfig",
    "SegmentationEfficientFormerModelConfig",
    "DetectionEfficientFormerModelConfig",
    "ClassificationMobileNetV3ModelConfig",
    "SegmentationMobileNetV3ModelConfig",
    "DetectionMobileNetV3ModelConfig",
    "ClassificationMobileViTModelConfig",
    "PIDNetModelConfig",
    "ClassificationResNetModelConfig",
    "SegmentationResNetModelConfig",
    "DetectionResNetModelConfig",
    "SegmentationSegFormerModelConfig",
    "ClassificationViTModelConfig",
    "DetectionYoloXModelConfig",
    "ClassificationMixNetSmallModelConfig",
    "ClassificationMixNetMediumModelConfig",
    "ClassificationMixNetLargeModelConfig",
    "SegmentationMixNetSmallModelConfig",
    "SegmentationMixNetMediumModelConfig",
    "SegmentationMixNetLargeModelConfig",
    "DetectionMixNetSmallModelConfig",
    "DetectionMixNetMediumModelConfig",
    "DetectionMixNetLargeModelConfig",
]


@dataclass
class ArchitectureConfig:
    full: Optional[Dict[str, Any]] = None
    backbone: Optional[Dict[str, Any]] = None
    neck: Optional[Dict[str, Any]] = None
    head: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        assert bool(self.full) != bool(self.backbone), "Only one of full or backbone should be given."

@dataclass
class CheckpointConfig:
    use_pretrained: bool = True
    load_head: bool = False
    path: Optional[Union[Path, str]] = None
    fx_model_path: Optional[Union[Path, str]] = None
    optimizer_path: Optional[Union[Path, str]] = None

@dataclass
class ModelConfig:
    task: str = MISSING
    name: str = MISSING
    checkpoint: CheckpointConfig = field(default_factory=lambda: CheckpointConfig())
    load_checkpoint_head: bool = False
    fx_model_checkpoint: Optional[Union[Path, str]] = None
    resume_optimizer_checkpoint: Optional[Union[Path, str]] = None
    freeze_backbone: bool = False
    architecture: ArchitectureConfig = field(default_factory=lambda: ArchitectureConfig())
    losses: Optional[List[Dict[str, Any]]] = None


@dataclass
class EfficientFormerArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "efficientformer",
        "params": {
            "num_attention_heads": 8,
            "attention_channels": 256,
            "attention_dropout_prob": 0.,
            "attention_value_expansion_ratio": 4,
            "ffn_intermediate_ratio": 4,
            "ffn_dropout_prob": 0.,
            "ffn_act_type": 'gelu',
            "vit_num": 1,
        },
        "stage_params": [
            {"num_blocks": 3, "channels": 48},
            {"num_blocks": 2, "channels": 96},
            {"num_blocks": 6, "channels": 224},
            {"num_blocks": 4, "channels": 448},
        ],
    })


@dataclass
class MobileNetV3ArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "mobilenetv3",
        "params": None,
        "stage_params": [
            {
                "in_channels": [16],
                "kernel_sizes": [3],
                "expanded_channels": [16],
                "out_channels": [16],
                "use_se": [True],
                "act_type": ["relu"],
                "stride": [2],
            },
            {
                "in_channels": [16, 24],
                "kernel_sizes": [3, 3],
                "expanded_channels": [72, 88],
                "out_channels": [24, 24],
                "use_se": [False, False],
                "act_type": ["relu", "relu"],
                "stride": [2, 1],
            },
            {
                "in_channels": [24, 40, 40, 40, 48],
                "kernel_sizes": [5, 5, 5, 5, 5],
                "expanded_channels": [96, 240, 240, 120, 144],
                "out_channels": [40, 40, 40, 48, 48],
                "use_se": [True, True, True, True, True],
                "act_type": ["hard_swish", "hard_swish", "hard_swish", "hard_swish", "hard_swish"],
                "stride": [2, 1, 1, 1, 1],
            },
            {
                "in_channels": [48, 96, 96],
                "kernel_sizes": [5, 5, 5],
                "expanded_channels": [288, 576, 576],
                "out_channels": [96, 96, 96],
                "use_se": [True, True, True],
                "act_type": ["hard_swish", "hard_swish", "hard_swish"],
                "stride": [2, 1, 1],
            },
        ],
    })


@dataclass
class MobileViTArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "mobilevit",
        "params": {
            "patch_size": 2,
            "num_attention_heads": 4,
            "attention_dropout_prob": 0.1,
            "ffn_dropout_prob": 0.0,
            "output_expansion_ratio": 4,
            "use_fusion_layer": True,
        },
        "stage_params": [
            {
                "out_channels": 32,
                "block_type": "mv2",
                "num_blocks": 1,
                "stride": 1,
                "ir_expansion_ratio": 4,
            },
            {
                "out_channels": 64,
                "block_type": "mv2",
                "num_blocks": 3,
                "stride": 2,
                "ir_expansion_ratio": 4,
            },
            {
                "out_channels": 96,
                "block_type": "mobilevit",
                "num_blocks": 2,
                "stride": 2,
                "hidden_size": 144,
                "intermediate_size": 288,
                "dilate": False,
                "ir_expansion_ratio": 4,
            },
            {
                "out_channels": 128,
                "block_type": "mobilevit",
                "num_blocks": 4,
                "stride": 2,
                "hidden_size": 192,
                "intermediate_size": 384,
                "dilate": False,
                "ir_expansion_ratio": 4,
            },
            {
                "out_channels": 160,
                "block_type": "mobilevit",
                "num_blocks": 3,
                "stride": 2,
                "hidden_size": 240,
                "intermediate_size": 480,
                "dilate": False,
                "ir_expansion_ratio": 4,
            },
        ]
    })


@dataclass
class PIDNetArchitectureConfig(ArchitectureConfig):
    full: Dict[str, Any] = field(default_factory=lambda: {
        "name": "pidnet",
        "m": 2,
        "n": 3,
        "channels": 32,
        "ppm_channels": 96,
        "head_channels": 128,
    })


@dataclass
class ResNetArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "resnet",
        "params": {
            "block_type": "bottleneck",
            "norm_type": "batch_norm",
        },
        "stage_params": [
            {"channels": 64, "num_blocks": 3},
            {"channels": 128, "num_blocks": 4, "replace_stride_with_dilation": False},
            {"channels": 256, "num_blocks": 6, "replace_stride_with_dilation": False},
            {"channels": 512, "num_blocks": 3, "replace_stride_with_dilation": False},
        ],
    })


@dataclass
class SegFormerArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "mixtransformer",
        "params": {
            "ffn_intermediate_expansion_ratio": 4,
            "ffn_act_type": "gelu",
            "ffn_dropout_prob": 0.0,
            "attention_dropout_prob": 0.0,
        },
        "stage_params": [
            {
                "num_blocks": 2,
                "sequence_reduction_ratio": 8,
                "attention_chananels": 32,
                "embedding_patch_sizes": 7,
                "embedding_strides": 4,
                "num_attention_heads": 1,
            },
            {
                "num_blocks": 2,
                "sequence_reduction_ratio": 4,
                "attention_chananels": 64,
                "embedding_patch_sizes": 3,
                "num_attention_heads": 2,
            },
            {
                "num_blocks": 2,
                "sequence_reduction_ratio": 2,
                "attention_chananels": 160,
                "embedding_patch_sizes": 3,
                "embedding_strides": 2,
                "num_attention_heads": 5,
            },
            {
                "num_blocks": 2,
                "sequence_reduction_ratio": 1,
                "attention_chananels": 256,
                "embedding_patch_sizes": 3,
                "embedding_strides": 2,
                "num_attention_heads": 8,
            },
        ],
    })


@dataclass
class ViTArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "vit",
        "params": {
            "patch_size": 16,
            "attention_channels": 192,
            "num_blocks": 12,
            "num_attention_heads": 3,
            "attention_dropout_prob": 0.0,
            "ffn_intermediate_channels": 768,
            "ffn_dropout_prob": 0.1,
            "use_cls_token": True,
            "vocab_size": 1000,
        },
        "stage_params": None,
    })


@dataclass
class MixNetSmallArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "mixnet",
        "params": {
            "stem_channels": 16,
            "wid_mul": 1.0,
            "dep_mul": 1.0,
            "dropout_rate": 0.,
        },
        "stage_params":  [
            {
                "expansion_ratio": [1, 6, 3],
                "out_channels": [16, 24, 24],
                "num_blocks": [1, 1, 1],
                "kernel_sizes": [[3], [3], [3]],
                "num_exp_groups": [1, 2, 2],
                "num_poi_groups": [1, 2, 2],
                "stride": [1, 2, 1],
                "act_type": ["relu", "relu", "relu"],
                "se_reduction_ratio": [None, None, None],
            },
            {
                "expansion_ratio": [6, 6],
                "out_channels": [40, 40],
                "num_blocks": [1, 3],
                "kernel_sizes": [[3, 5, 7], [3, 5]],
                "num_exp_groups": [1, 2],
                "num_poi_groups": [1, 2],
                "stride": [2, 1],
                "act_type": ["swish", "swish"],
                "se_reduction_ratio": [2, 2],
            },
            {
                "expansion_ratio": [6, 6, 6, 3],
                "out_channels": [80, 80, 120, 120],
                "num_blocks": [1, 2, 1, 2],
                "kernel_sizes": [[3, 5, 7], [3, 5], [3, 5, 7], [3, 5, 7, 9]],
                "num_exp_groups": [1, 1, 2, 2],
                "num_poi_groups": [2, 2, 2, 2],
                "stride": [2, 1, 1, 1],
                "act_type": ["swish", "swish", "swish", "swish"],
                "se_reduction_ratio": [4, 4, 2, 2],
            },
            {
                "expansion_ratio": [6, 6],
                "out_channels": [200, 200],
                "num_blocks": [1, 2],
                "kernel_sizes": [[3, 5, 7, 9, 11], [3, 5, 7, 9]],
                "num_exp_groups": [1, 1],
                "num_poi_groups": [1, 2],
                "stride": [2, 1],
                "act_type": ["swish", "swish"],
                "se_reduction_ratio": [2, 2],
            },
        ],
    })


@dataclass
class MixNetMediumArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "mixnet",
        "params": {
            "stem_channels": 24,
            "wid_mul": 1.0,
            "dep_mul": 1.0,
            "dropout_rate": 0.,
        },
        "stage_params":  [
            {
                "expansion_ratio": [1, 6, 3],
                "out_channels": [24, 32, 32],
                "num_blocks": [1, 1, 1],
                "kernel_sizes": [[3], [3, 5, 7], [3]],
                "num_exp_groups": [1, 2, 2],
                "num_poi_groups": [1, 2, 2],
                "stride": [1, 2, 1],
                "act_type": ["relu", "relu", "relu"],
                "se_reduction_ratio": [None, None, None],
            },
            {
                "expansion_ratio": [6, 6],
                "out_channels": [40, 40],
                "num_blocks": [1, 3],
                "kernel_sizes": [[3, 5, 7, 9], [3, 5]],
                "num_exp_groups": [1, 2],
                "num_poi_groups": [1, 2],
                "stride": [2, 1],
                "act_type": ["swish", "swish"],
                "se_reduction_ratio": [2, 2],
            },
            {
                "expansion_ratio": [6, 6, 6, 3],
                "out_channels": [80, 80, 120, 120],
                "num_blocks": [1, 3, 1, 3],
                "kernel_sizes": [[3, 5, 7], [3, 5, 7, 9], [3], [3, 5, 7, 9]],
                "num_exp_groups": [1, 2, 1, 2],
                "num_poi_groups": [1, 2, 1, 2],
                "stride": [2, 1, 1, 1],
                "act_type": ["swish", "swish", "swish", "swish"],
                "se_reduction_ratio": [4, 4, 2, 2],
            },
            {
                "expansion_ratio": [6, 6],
                "out_channels": [200, 200],
                "num_blocks": [1, 3],
                "kernel_sizes": [[3, 5, 7, 9], [3, 5, 7, 9]],
                "num_exp_groups": [1, 1],
                "num_poi_groups": [1, 2],
                "stride": [2, 1],
                "act_type": ["swish", "swish"],
                "se_reduction_ratio": [2, 2],
            },
        ],
    })


@dataclass
class MixNetLargeArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "mixnet",
        "params": {
            "stem_channels": 24,
            "wid_mul": 1.3,
            "dep_mul": 1.0,
            "dropout_rate": 0.,
        },
        "stage_params":  [
            {
                "expansion_ratio": [1, 6, 3],
                "out_channels": [24, 32, 32],
                "num_blocks": [1, 1, 1],
                "kernel_sizes": [[3], [3, 5, 7], [3]],
                "num_exp_groups": [1, 2, 2],
                "num_poi_groups": [1, 2, 2],
                "stride": [1, 2, 1],
                "act_type": ["relu", "relu", "relu"],
                "se_reduction_ratio": [None, None, None],
            },
            {
                "expansion_ratio": [6, 6],
                "out_channels": [40, 40],
                "num_blocks": [1, 3],
                "kernel_sizes": [[3, 5, 7, 9], [3, 5]],
                "num_exp_groups": [1, 2],
                "num_poi_groups": [1, 2],
                "stride": [2, 1],
                "act_type": ["swish", "swish"],
                "se_reduction_ratio": [2, 2],
            },
            {
                "expansion_ratio": [6, 6, 6, 3],
                "out_channels": [80, 80, 120, 120],
                "num_blocks": [1, 3, 1, 3],
                "kernel_sizes": [[3, 5, 7], [3, 5, 7, 9], [3], [3, 5, 7, 9]],
                "num_exp_groups": [1, 2, 1, 2],
                "num_poi_groups": [1, 2, 1, 2],
                "stride": [2, 1, 1, 1],
                "act_type": ["swish", "swish", "swish", "swish"],
                "se_reduction_ratio": [4, 4, 2, 2],
            },
            {
                "expansion_ratio": [6, 6],
                "out_channels": [200, 200],
                "num_blocks": [1, 3],
                "kernel_sizes": [[3, 5, 7, 9], [3, 5, 7, 9]],
                "num_exp_groups": [1, 1],
                "num_poi_groups": [1, 2],
                "stride": [2, 1],
                "act_type": ["swish", "swish"],
                "se_reduction_ratio": [2, 2],
            },
        ],
    })


@dataclass
class CSPDarkNetSmallArchitectureConfig(ArchitectureConfig):
    backbone: Dict[str, Any] = field(default_factory=lambda: {
        "name": "cspdarknet",
        "params": {
            "dep_mul": 0.33,
            "wid_mul": 0.5,
            "act_type": "silu",
        },
        "stage_params": None,
    })


@dataclass
class ClassificationEfficientFormerModelConfig(ModelConfig):
    task: str = "classification"
    name: str = "efficientformer_l1"
    architecture: ArchitectureConfig = field(default_factory=lambda: EfficientFormerArchitectureConfig(
        head={
            "name": "fc",
            "params": {
                "intermediate_channels": 1024,
                "num_layers": 1,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "label_smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationEfficientFormerModelConfig(ModelConfig):
    task: str = "segmentation"
    name: str = "efficientformer_l1"
    architecture: ArchitectureConfig = field(default_factory=lambda: EfficientFormerArchitectureConfig(
        head={
            "name": "all_mlp_decoder",
            "params": {
                "intermediate_channels": 256,
                "classifier_dropout_prob": 0.,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class DetectionEfficientFormerModelConfig(ModelConfig):
    task: str = "detection"
    name: str = "efficientformer_l1"
    checkpoint: CheckpointConfig = field(default_factory=lambda: CheckpointConfig(
        load_head=True
    ))
    architecture: ArchitectureConfig = field(default_factory=lambda: EfficientFormerArchitectureConfig(
        neck={
            "name": "fpn",
            "params": {
                "num_outs": 4,
                "start_level": 0,
                "end_level": -1,
                "add_extra_convs": False,
                "relu_before_extra_convs": False,
            },
        },
        head={
            "name": "anchor_decoupled_head",
            "params": {
                # Anchor parameters
                "anchor_sizes": [[32,], [64,], [128,], [256,]],
                "aspect_ratios": [0.5, 1.0, 2.0],
                "num_layers": 1,
                "norm_type": "batch_norm",
                # postprocessor - decode
                "topk_candidates": 1000,
                "score_thresh": 0.05,
                # postprocessor - nms
                "nms_thresh": 0.45,
                "class_agnostic": False,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "retinanet_loss", "weight": None},
    ])


@dataclass
class ClassificationMobileNetV3ModelConfig(ModelConfig):
    task: str = "classification"
    name: str = "mobilenet_v3_small"
    architecture: ArchitectureConfig = field(default_factory=lambda: MobileNetV3ArchitectureConfig(
        head={
            "name": "fc",
            "params": {
                "intermediate_channels": 1024,
                "num_layers": 1,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "label_smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationMobileNetV3ModelConfig(ModelConfig):
    task: str = "segmentation"
    name: str = "mobilenet_v3_small"
    architecture: ArchitectureConfig = field(default_factory=lambda: MobileNetV3ArchitectureConfig(
        head={
            "name": "all_mlp_decoder",
            "params": {
                "intermediate_channels": 256,
                "classifier_dropout_prob": 0.,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class DetectionMobileNetV3ModelConfig(ModelConfig):
    task: str = "detection"
    name: str = "mobilenet_v3_small"
    checkpoint: CheckpointConfig = field(default_factory=lambda: CheckpointConfig(
        load_head=True
    ))
    architecture: ArchitectureConfig = field(default_factory=lambda: MobileNetV3ArchitectureConfig(
        neck={
            "name": "fpn",
            "params": {
                "num_outs": 4,
                "start_level": 0,
                "end_level": -1,
                "add_extra_convs": False,
                "relu_before_extra_convs": False,
            },
        },
        head={
            "name": "anchor_decoupled_head",
            "params": {
                # Anchor parameters
                "anchor_sizes": [[32,], [64,], [128,], [256,]],
                "aspect_ratios": [0.5, 1.0, 2.0],
                "num_layers": 1,
                "norm_type": "batch_norm",
                # postprocessor - decode
                "topk_candidates": 1000,
                "score_thresh": 0.05,
                # postprocessor - nms
                "nms_thresh": 0.45,
                "class_agnostic": False,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "retinanet_loss", "weight": None},
    ])


@dataclass
class ClassificationMobileViTModelConfig(ModelConfig):
    task: str = "classification"
    name: str = "mobilevit_s"
    architecture: ArchitectureConfig = field(default_factory=lambda: MobileViTArchitectureConfig(
        head={
            "name": "fc",
            "params": {
                "intermediate_channels": 1024,
                "num_layers": 1,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "label_smoothing": 0.1, "weight": None}
    ])


@dataclass
class PIDNetModelConfig(ModelConfig):
    task: str = "segmentation"
    name: str = "pidnet_s"
    architecture: ArchitectureConfig = field(default_factory=lambda: PIDNetArchitectureConfig())
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "pidnet_loss", "ignore_index": 255, "weight": None},
    ])


@dataclass
class ClassificationResNetModelConfig(ModelConfig):
    task: str = "classification"
    name: str = "resnet50"
    architecture: ArchitectureConfig = field(default_factory=lambda: ResNetArchitectureConfig(
        head={
            "name": "fc",
            "params": {
                "intermediate_channels": 1024,
                "num_layers": 1,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "label_smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationResNetModelConfig(ModelConfig):
    task: str = "segmentation"
    name: str = "resnet50"
    architecture: ArchitectureConfig = field(default_factory=lambda: ResNetArchitectureConfig(
        head={
            "name": "all_mlp_decoder",
            "params": {
                "intermediate_channels": 256,
                "classifier_dropout_prob": 0.,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class DetectionResNetModelConfig(ModelConfig):
    task: str = "detection"
    name: str = "resnet50"
    checkpoint: CheckpointConfig = field(default_factory=lambda: CheckpointConfig(
        load_head=True
    ))
    architecture: ArchitectureConfig = field(default_factory=lambda: ResNetArchitectureConfig(
        neck={
            "name": "fpn",
            "params": {
                "num_outs": 4,
                "start_level": 0,
                "end_level": -1,
                "add_extra_convs": False,
                "relu_before_extra_convs": False,
            },
        },
        head={
            "name": "anchor_decoupled_head",
            "params": {
                # Anchor parameters
                "anchor_sizes": [[32,], [64,], [128,], [256,]],
                "aspect_ratios": [0.5, 1.0, 2.0],
                "num_layers": 1,
                "norm_type": "batch_norm",
                # postprocessor - decode
                "topk_candidates": 1000,
                "score_thresh": 0.05,
                # postprocessor - nms
                "nms_thresh": 0.45,
                "class_agnostic": False,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "retinanet_loss", "weight": None},
    ])


@dataclass
class SegmentationSegFormerModelConfig(ModelConfig):
    task: str = "segmentation"
    name: str = "segformer"
    architecture: ArchitectureConfig = field(default_factory=lambda: SegFormerArchitectureConfig(
        head={
            "name": "all_mlp_decoder",
            "params": {
                "intermediate_channels": 256,
                "classifier_dropout_prob": 0.,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class ClassificationViTModelConfig(ModelConfig):
    task: str = "classification"
    name: str = "vit_tiny"
    architecture: ArchitectureConfig = field(default_factory=lambda: ViTArchitectureConfig(
        head={
            "name": "fc",
            "params": {
                "intermediate_channels": 1024,
                "num_layers": 1,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "label_smoothing": 0.1, "weight": None}
    ])


@dataclass
class DetectionYoloXModelConfig(ModelConfig):
    task: str = "detection"
    name: str = "yolox_s"
    checkpoint: CheckpointConfig = field(default_factory=lambda: CheckpointConfig(
        load_head=True
    ))
    architecture: ArchitectureConfig = field(default_factory=lambda: CSPDarkNetSmallArchitectureConfig(
        neck={
            "name": "yolopafpn",
            "params": {
                "dep_mul": 0.33,
                "act_type": "silu",
            },
        },
        head={
            "name": "anchor_free_decoupled_head",
            "params": {
                "act_type": "silu",
                # postprocessor - decode
                "score_thresh": 0.7,
                # postprocessor - nms
                "nms_thresh": 0.45,
                "class_agnostic": False,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "yolox_loss", "weight": None}
    ])


@dataclass
class ClassificationMixNetSmallModelConfig(ModelConfig):
    task: str = "classification"
    name: str = "mixnet_s"
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetSmallArchitectureConfig(
        head={
            "name": "fc",
            "params": {
                "intermediate_channels": 1024,
                "num_layers": 1,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "label_smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationMixNetSmallModelConfig(ModelConfig):
    task: str = "segmentation"
    name: str = "mixnet_s"
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetSmallArchitectureConfig(
        head={
            "name": "all_mlp_decoder",
            "params": {
                "intermediate_channels": 256,
                "classifier_dropout_prob": 0.,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class DetectionMixNetSmallModelConfig(ModelConfig):
    task: str = "detection"
    name: str = "mixnet_s"
    checkpoint: CheckpointConfig = field(default_factory=lambda: CheckpointConfig(
        load_head=True
    ))
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetSmallArchitectureConfig(
        neck={
            "name": "fpn",
            "params": {
                "num_outs": 4,
                "start_level": 0,
                "end_level": -1,
                "add_extra_convs": False,
                "relu_before_extra_convs": False,
            },
        },
        head={
            "name": "anchor_decoupled_head",
            "params": {
                # Anchor parameters
                "anchor_sizes": [[32,], [64,], [128,], [256,]],
                "aspect_ratios": [0.5, 1.0, 2.0],
                "num_layers": 1,
                "norm_type": "batch_norm",
                # postprocessor - decode
                "topk_candidates": 1000,
                "score_thresh": 0.05,
                # postprocessor - nms
                "nms_thresh": 0.45,
                "class_agnostic": False,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "retinanet_loss", "weight": None},
    ])


@dataclass
class ClassificationMixNetMediumModelConfig(ModelConfig):
    task: str = "classification"
    name: str = "mixnet_m"
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetMediumArchitectureConfig(
        head={
            "name": "fc",
            "params": {
                "intermediate_channels": 1024,
                "num_layers": 1,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "label_smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationMixNetMediumModelConfig(ModelConfig):
    task: str = "segmentation"
    name: str = "mixnet_m"
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetMediumArchitectureConfig(
        head={
            "name": "all_mlp_decoder",
            "params": {
                "intermediate_channels": 256,
                "classifier_dropout_prob": 0.,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class DetectionMixNetMediumModelConfig(ModelConfig):
    task: str = "detection"
    name: str = "mixnet_m"
    checkpoint: CheckpointConfig = field(default_factory=lambda: CheckpointConfig(
        load_head=True
    ))
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetMediumArchitectureConfig(
        neck={
            "name": "fpn",
            "params": {
                "num_outs": 4,
                "start_level": 0,
                "end_level": -1,
                "add_extra_convs": False,
                "relu_before_extra_convs": False,
            },
        },
        head={
            "name": "anchor_decoupled_head",
            "params": {
                # Anchor parameters
                "anchor_sizes": [[32,], [64,], [128,], [256,]],
                "aspect_ratios": [0.5, 1.0, 2.0],
                "num_layers": 1,
                "norm_type": "batch_norm",
                # postprocessor - decode
                "topk_candidates": 1000,
                "score_thresh": 0.05,
                # postprocessor - nms
                "nms_thresh": 0.45,
                "class_agnostic": False,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "retinanet_loss", "weight": None},
    ])


@dataclass
class ClassificationMixNetLargeModelConfig(ModelConfig):
    task: str = "classification"
    name: str = "mixnet_l"
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetLargeArchitectureConfig(
        head={
            "name": "fc",
            "params": {
                "intermediate_channels": 1024,
                "num_layers": 1,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "label_smoothing": 0.1, "weight": None}
    ])


@dataclass
class SegmentationMixNetLargeModelConfig(ModelConfig):
    task: str = "segmentation"
    name: str = "mixnet_l"
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetLargeArchitectureConfig(
        head={
            "name": "all_mlp_decoder",
            "params": {
                "intermediate_channels": 256,
                "classifier_dropout_prob": 0.,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "cross_entropy", "ignore_index": 255, "weight": None}
    ])


@dataclass
class DetectionMixNetLargeModelConfig(ModelConfig):
    task: str = "detection"
    name: str = "mixnet_l"
    checkpoint: CheckpointConfig = field(default_factory=lambda: CheckpointConfig(
        load_head=True
    ))
    architecture: ArchitectureConfig = field(default_factory=lambda: MixNetLargeArchitectureConfig(
        neck={
            "name": "fpn",
            "params": {
                "num_outs": 4,
                "start_level": 0,
                "end_level": -1,
                "add_extra_convs": False,
                "relu_before_extra_convs": False,
            },
        },
        head={
            "name": "anchor_decoupled_head",
            "params": {
                # Anchor parameters
                "anchor_sizes": [[32,], [64,], [128,], [256,]],
                "aspect_ratios": [0.5, 1.0, 2.0],
                "num_layers": 1,
                "norm_type": "batch_norm",
                # postprocessor - decode
                "topk_candidates": 1000,
                "score_thresh": 0.05,
                # postprocessor - nms
                "nms_thresh": 0.45,
                "class_agnostic": False,
            }
        }
    ))
    losses: List[Dict[str, Any]] = field(default_factory=lambda: [
        {"criterion": "retinanet_loss", "weight": None},
    ])



import json
import re
from itertools import repeat


def read_json(json_path):
    with open(json_path, 'r') as f:
        data = json.load(f)
    return data

def expand_to_chs(x, n):
    if not isinstance(x, (tuple, list)):
        x = tuple(repeat(x, n))
    elif len(x) == 1:
        x = x * n
    else:
        assert len(x) == n, 'normalization stats must match image channels'
    return x

def natural_key(string_):
    """See http://www.codinghorror.com/blog/archives/001018.html"""
    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_.lower())]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# Copyright (c) Megvii Inc. All rights reserved.
"""
Based on the YOLOX implementation of Megvii.
https://github.com/Megvii-BaseDetection/YOLOX/blob/main/yolox/models/yolo_head.py
"""
from typing import List

from omegaconf import DictConfig
import torch
import torch.nn as nn

from ....op.custom import ConvLayer
from ....utils import ModelOutput


class AnchorFreeDecoupledHead(nn.Module):
    def __init__(
        self,
        num_classes: int,
        intermediate_features_dim: List[int],
        params: DictConfig,
    ):
        super().__init__()
        act_type = params.act_type

        self.num_classes = num_classes

        self.cls_convs = nn.ModuleList()
        self.reg_convs = nn.ModuleList()
        self.cls_preds = nn.ModuleList()
        self.reg_preds = nn.ModuleList()
        self.obj_preds = nn.ModuleList()
        self.stems = nn.ModuleList()
        Conv = ConvLayer

        hidden_dim = int(intermediate_features_dim[0]) # 256 * width
        for i in range(len(intermediate_features_dim)):
            self.stems.append(
                Conv(
                    in_channels=int(intermediate_features_dim[i]),
                    out_channels=hidden_dim,
                    kernel_size=1,
                    stride=1,
                    act_type=act_type,
                )
            )
            self.cls_convs.append(
                nn.Sequential(
                    *[
                        Conv(
                            in_channels=hidden_dim,
                            out_channels=hidden_dim,
                            kernel_size=3,
                            stride=1,
                            act_type=act_type,
                        ),
                        Conv(
                            in_channels=hidden_dim,
                            out_channels=hidden_dim,
                            kernel_size=3,
                            stride=1,
                            act_type=act_type,
                        ),
                    ]
                )
            )
            self.reg_convs.append(
                nn.Sequential(
                    *[
                        Conv(
                            in_channels=hidden_dim,
                            out_channels=hidden_dim,
                            kernel_size=3,
                            stride=1,
                            act_type=act_type,
                        ),
                        Conv(
                            in_channels=hidden_dim,
                            out_channels=hidden_dim,
                            kernel_size=3,
                            stride=1,
                            act_type=act_type,
                        ),
                    ]
                )
            )
            self.cls_preds.append(
                nn.Conv2d(
                    in_channels=hidden_dim,
                    out_channels=self.num_classes,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
            )
            self.reg_preds.append(
                nn.Conv2d(
                    in_channels=hidden_dim,
                    out_channels=4,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
            )
            self.obj_preds.append(
                nn.Conv2d(
                    in_channels=hidden_dim,
                    out_channels=1,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
            )

    def forward(self, xin):
        outputs = []

        for k, (cls_conv, reg_conv, x) in enumerate(zip(self.cls_convs, self.reg_convs, xin)):
            x = self.stems[k](x)
            cls_x = x
            reg_x = x

            cls_feat = cls_conv(cls_x)
            cls_output = self.cls_preds[k](cls_feat)

            reg_feat = reg_conv(reg_x)
            reg_output = self.reg_preds[k](reg_feat)
            obj_output = self.obj_preds[k](reg_feat)

            output = torch.cat([reg_output, obj_output, cls_output], 1)

            outputs.append(output)

        return ModelOutput(pred=outputs)


def anchor_free_decoupled_head(num_classes, intermediate_features_dim, conf_model_head, **kwargs) -> AnchorFreeDecoupledHead:
    return AnchorFreeDecoupledHead(num_classes=num_classes,
                                   intermediate_features_dim=intermediate_features_dim,
                                   params=conf_model_head.params)



from typing import Callable, Dict

from .custom import (
    AutoAugment,
    CenterCrop,
    ColorJitter,
    Pad,
    RandomCrop,
    RandomCutmix,
    RandomErasing,
    RandomHorizontalFlip,
    RandomMixup,
    RandomResizedCrop,
    RandomVerticalFlip,
    Resize,
    TrivialAugmentWide,
)

TRANSFORM_DICT: Dict[str, Callable] = {
    'centercrop': CenterCrop,
    'colorjitter': ColorJitter,
    'pad': Pad,
    'randomcrop': RandomCrop,
    'randomresizedcrop': RandomResizedCrop,
    'randomhorizontalflip': RandomHorizontalFlip,
    'randomverticalflip': RandomVerticalFlip,
    'randomerasing': RandomErasing,
    'resize': Resize,
    'mixup': RandomMixup,
    'cutmix': RandomCutmix,
    'trivialaugmentwide': TrivialAugmentWide,
    'autoaugment': AutoAugment,
}



import os
from pathlib import Path

from netspresso_trainer.trainer_util import parse_args_netspresso, train_with_yaml_impl

LOG_LEVEL = os.getenv('LOG_LEVEL', 'INFO')

def train_cli() -> None:
    args_parsed = parse_args_netspresso(with_gpus=True)

    logging_dir: Path = train_with_yaml_impl(
        gpus=args_parsed.gpus,
        data=args_parsed.data,
        augmentation=args_parsed.augmentation,
        model=args_parsed.model,
        training=args_parsed.training,
        logging=args_parsed.logging,
        environment=args_parsed.environment,
        log_level=args_parsed.log_level
    )

    return logging_dir



from typing import Any, Optional

import torch

from ..models.utils import ModelOutput


class SegmentationPostprocessor:
    def __init__(self, conf_model):
        pass

    def __call__(self, outputs: ModelOutput):
        pred = outputs['pred']
        pred = torch.max(pred, dim=1)[1]  # argmax
        return pred



import os
from functools import partial
from pathlib import Path
from typing import Dict, List, Optional, Type, Union

import torch.distributed as dist
from loguru import logger

from .augmentation.registry import TRANSFORM_DICT
from .classification import classification_mix_collate_fn, classification_onehot_collate_fn
from .detection import detection_collate_fn
from .registry import CREATE_TRANSFORM, CUSTOM_DATASET, DATA_SAMPLER, HUGGINGFACE_DATASET
from .utils.loader import create_loader

TRAIN_VALID_SPLIT_RATIO = 0.9

def build_dataset(conf_data, conf_augmentation, task: str, model_name: str, distributed: bool):

    if not distributed or dist.get_rank() == 0:
        logger.info('-'*40)
        logger.info("Loading data...")

    task = conf_data.task

    assert task in DATA_SAMPLER, f"Data sampler for {task} is not yet supported!"

    train_transform = CREATE_TRANSFORM(model_name, is_training=True)
    target_transform = CREATE_TRANSFORM(model_name, is_training=False)

    data_format = conf_data.format

    assert data_format in ['local', 'huggingface'], f"No such data format named {data_format} in {['local', 'huggingface']}!"

    if data_format == 'local':
        assert task in CUSTOM_DATASET, f"Local dataset for {task} is not yet supported!"
        data_sampler = DATA_SAMPLER[task](conf_data, train_valid_split_ratio=TRAIN_VALID_SPLIT_RATIO)

        train_samples, valid_samples, test_samples, misc = data_sampler.load_samples()
        idx_to_class = misc['idx_to_class'] if 'idx_to_class' in misc else None
        label_value_to_idx = misc['label_value_to_idx'] if 'label_value_to_idx' in misc else None
        test_with_label = misc['test_with_label'] if 'test_with_label' in misc else None

        train_dataset = CUSTOM_DATASET[task](
            conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='train',
            samples=train_samples, transform=train_transform, label_value_to_idx=label_value_to_idx
        )

        valid_dataset = None
        if valid_samples is not None:
            valid_dataset = CUSTOM_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='valid',
                samples=valid_samples, transform=target_transform, label_value_to_idx=label_value_to_idx
            )

        test_dataset = None
        if test_samples is not None:
            test_dataset = CUSTOM_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='test',
                samples=test_samples, transform=target_transform,
                with_label=test_with_label, label_value_to_idx=label_value_to_idx
            )

    elif data_format == 'huggingface':
        assert task in CUSTOM_DATASET, f"HuggingFace dataset for {task} is not yet supported!"
        assert task in DATA_SAMPLER, f"Data sampler for {task} is not yet supported!"

        data_sampler = DATA_SAMPLER[task](conf_data, train_valid_split_ratio=TRAIN_VALID_SPLIT_RATIO)

        train_samples, valid_samples, test_samples, misc = data_sampler.load_huggingface_samples()
        idx_to_class = misc['idx_to_class'] if 'idx_to_class' in misc else None
        label_value_to_idx = misc['label_value_to_idx'] if 'label_value_to_idx' in misc else None
        test_with_label = misc['test_with_label'] if 'test_with_label' in misc else None

        train_dataset = HUGGINGFACE_DATASET[task](
            conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='train',
            huggingface_dataset=train_samples, transform=train_transform, label_value_to_idx=label_value_to_idx
        )

        valid_dataset = None
        if valid_samples is not None:
            valid_dataset = HUGGINGFACE_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='valid',
                huggingface_dataset=valid_samples, transform=target_transform, label_value_to_idx=label_value_to_idx
            )

        test_dataset = None
        if test_samples is not None:
            test_dataset = HUGGINGFACE_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='test',
                huggingface_dataset=test_samples, transform=target_transform, label_value_to_idx=label_value_to_idx
            )

    if not distributed or dist.get_rank() == 0:
        logger.info(f"Summary | Dataset: <{conf_data.name}> (with {data_format} format)")
        logger.info(f"Summary | Training dataset: {len(train_dataset)} sample(s)")
        if valid_dataset is not None:
            logger.info(f"Summary | Validation dataset: {len(valid_dataset)} sample(s)")
        if test_dataset is not None:
            logger.info(f"Summary | Test dataset: {len(test_dataset)} sample(s)")

    return train_dataset, valid_dataset, test_dataset


def build_dataloader(conf, task: str, model_name: str, train_dataset, eval_dataset, profile=False):

    if task == 'classification':
        conf_mix_transform = getattr(conf.augmentation, 'mix_transforms', None)
        if conf_mix_transform:
            mix_transforms = []
            for mix_transform_conf in conf.augmentation.mix_transforms:
                name = mix_transform_conf.name.lower()

                mix_kwargs = list(mix_transform_conf.keys())
                mix_kwargs.remove('name')
                mix_kwargs = {k:mix_transform_conf[k] for k in mix_kwargs}
                mix_kwargs['num_classes'] = train_dataset.num_classes

                transform = TRANSFORM_DICT[name](**mix_kwargs)
                mix_transforms.append(transform)

            train_collate_fn = partial(classification_mix_collate_fn, mix_transforms=mix_transforms)
            eval_collate_fn = partial(classification_onehot_collate_fn, num_classes=train_dataset.num_classes)
        else:
            train_collate_fn = None
            eval_collate_fn = None

        train_loader = create_loader(
            train_dataset,
            conf.data.name,
            logger,
            input_size=conf.augmentation.img_size,
            batch_size=conf.training.batch_size,
            is_training=True,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=train_collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            kwargs=None
        )

        eval_loader = create_loader(
            eval_dataset,
            conf.data.name,
            logger,
            input_size=conf.augmentation.img_size,
            batch_size=conf.training.batch_size,
            is_training=False,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=eval_collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            kwargs=None
        )
    elif task == 'segmentation':
        collate_fn = None

        train_loader = create_loader(
            train_dataset,
            conf.data.name,
            logger,
            batch_size=conf.training.batch_size,
            is_training=True,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            kwargs=None
        )

        eval_loader = create_loader(
            eval_dataset,
            conf.data.name,
            logger,
            batch_size=conf.training.batch_size if model_name == 'pidnet' and not conf.distributed else 1,
            is_training=False,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=None,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            kwargs=None
        )
    elif task == 'detection':
        collate_fn = detection_collate_fn

        train_loader = create_loader(
            train_dataset,
            conf.data.name,
            logger,
            batch_size=conf.training.batch_size,
            is_training=True,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            kwargs=None
        )

        eval_loader = create_loader(
            eval_dataset,
            conf.data.name,
            logger,
            # TODO: support batch size 1 inference
            batch_size=conf.training.batch_size if not conf.distributed else 2,
            is_training=False,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            kwargs=None
        )

    else:
        raise AssertionError(f"Task ({task}) is not understood!")

    return train_loader, eval_loader



import os
from typing import Literal

import torch
from loguru import logger
from omegaconf import OmegaConf

from .base import BasePipeline

MAX_SAMPLE_RESULT = 10


class ClassificationPipeline(BasePipeline):
    def __init__(self, conf, task, model_name, model, devices,
                 train_dataloader, eval_dataloader, class_map, logging_dir, **kwargs):
        super(ClassificationPipeline, self).__init__(conf, task, model_name, model, devices,
                                                     train_dataloader, eval_dataloader, class_map, logging_dir, **kwargs)

    def train_step(self, batch):
        self.model.train()
        images, labels = batch
        images = images.to(self.devices)
        labels = labels.to(self.devices)
        target = {'target': labels}

        self.optimizer.zero_grad()

        out = self.model(images)
        self.loss_factory.calc(out, target, phase='train')
        if labels.dim() > 1: # Soft label to label number
            labels = torch.argmax(labels, dim=-1)
        pred = self.postprocessor(out)
        self.metric_factory.calc(pred, labels, phase='train')

        self.loss_factory.backward()
        self.optimizer.step()

        if self.conf.distributed:
            torch.distributed.barrier()

    def valid_step(self, batch):
        self.model.eval()
        images, labels = batch
        images = images.to(self.devices)
        labels = labels.to(self.devices)
        target = {'target': labels}

        out = self.model(images)
        self.loss_factory.calc(out, target, phase='valid')
        if labels.dim() > 1: # Soft label to label number
            labels = torch.argmax(labels, dim=-1)
        pred = self.postprocessor(out)
        self.metric_factory.calc(pred, labels, phase='valid')

        if self.conf.distributed:
            torch.distributed.barrier()

    def test_step(self, batch):
        self.model.eval()
        images, _ = batch
        images = images.to(self.devices)

        out = self.model(images.unsqueeze(0))
        pred = self.postprocessor(out, k=1)

        if self.conf.distributed:
            torch.distributed.barrier()

        return pred

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid']):
        pass



import torch.optim as optim


class Adadelta(optim.Adadelta):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        rho = optimizer_conf.rho
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, rho=rho, weight_decay=weight_decay)


class Adagrad(optim.Adagrad):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        lr_decay = optimizer_conf.lr_decay
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, lr_decay=lr_decay, weight_decay=weight_decay)


class Adam(optim.Adam):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        betas = optimizer_conf.betas
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, betas=betas, weight_decay=weight_decay)


class Adamax(optim.Adamax):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        betas = optimizer_conf.betas
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, betas=betas, weight_decay=weight_decay)


class AdamW(optim.AdamW):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        betas = optimizer_conf.betas
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, betas=betas, weight_decay=weight_decay)


class RMSprop(optim.RMSprop):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        alpha = optimizer_conf.alpha
        weight_decay = optimizer_conf.weight_decay
        momentum = optimizer_conf.momentum
        eps = optimizer_conf.eps

        super().__init__(params=params, lr=lr, alpha=alpha, weight_decay=weight_decay, momentum=momentum, eps=eps)


class SGD(optim.SGD):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        momentum = optimizer_conf.momentum
        weight_decay = optimizer_conf.weight_decay
        nesterov = optimizer_conf.nesterov

        super().__init__(params=params, lr=lr, momentum=momentum, weight_decay=weight_decay, nesterov=nesterov)



import math
import random
from collections.abc import Sequence
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import PIL.Image as Image
import torch
import torchvision.transforms as T
import torchvision.transforms.functional as F
from omegaconf import ListConfig
from torch import Tensor
from torch.nn import functional as F_torch
from torchvision.transforms.autoaugment import _apply_op
from torchvision.transforms.functional import InterpolationMode

BBOX_CROP_KEEP_THRESHOLD = 0.2
MAX_RETRY = 5
INVERSE_MODES_MAPPING = {
    'nearest': InterpolationMode.NEAREST,
    'bilinear': InterpolationMode.BILINEAR,
    'bicubic': InterpolationMode.BICUBIC,
}


class Compose:
    def __init__(self, transforms, additional_targets: Dict = None):
        if additional_targets is None:
            additional_targets = {}
        self.transforms = transforms
        self.additional_targets = additional_targets

    def _get_transformed(self, image, mask, bbox, visualize_for_debug):
        for t in self.transforms:
            if visualize_for_debug and not t.visualize:
                continue
            image, mask, bbox = t(image=image, mask=mask, bbox=bbox)
        return image, mask, bbox

    def __call__(self, image, mask=None, bbox=None, visualize_for_debug=False, **kwargs):
        additional_targets_result = {k: None for k in kwargs if k in self.additional_targets}

        result_image, result_mask, result_bbox = self._get_transformed(image=image, mask=mask, bbox=bbox, visualize_for_debug=visualize_for_debug)
        for key in additional_targets_result:
            if self.additional_targets[key] == 'mask':
                _, additional_targets_result[key], _ = self._get_transformed(image=image, mask=kwargs[key], bbox=None, visualize_for_debug=visualize_for_debug)
            elif self.additional_targets[key] == 'bbox':
                _, _, additional_targets_result[key] = self._get_transformed(image=image, mask=None, bbox=kwargs[key], visualize_for_debug=visualize_for_debug)
            else:
                del additional_targets_result[key]

        return_dict = {'image': result_image}
        if mask is not None:
            return_dict.update({'mask': result_mask})
        if bbox is not None:
            return_dict.update({'bbox': result_bbox})
        return_dict.update(additional_targets_result)
        return return_dict

    def __repr__(self):
        compose_summary = "CustomCompose"
        compose_list = ",\n\t".join([str(t) for t in self.transforms])
        compose_summary += "(\n\t" + compose_list + "\n)"
        return compose_summary


class CenterCrop(T.CenterCrop):
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
    ):
        super().__init__(size)

    def forward(self, image, mask=None, bbox=None):
        # TODO: Compute mask, bbox
        return F.center_crop(image, self.size), mask, bbox

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(size={self.size})"


class Identity:
    visualize = True

    def __init__(self):
        pass

    def __call__(self, image, mask=None, bbox=None):
        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "()"


class Pad(T.Pad):
    visualize = True
    def __init__(
        self,
        padding: Union[int, List],
        fill: Union[int, List],
        padding_mode: str,
    ):
        super().__init__(padding, fill, padding_mode)

    def forward(self, image, mask=None, bbox=None):
        image = F.pad(image, self.padding, self.fill, self.padding_mode)
        if mask is not None:
            mask = F.pad(mask, self.padding, fill=255, padding_mode=self.padding_mode)
        if bbox is not None:
            target_padding = [self.padding] if not isinstance(self.padding, Sequence) else self.padding

            padding_left, padding_top, _, _ = \
                target_padding * (4 / len(target_padding))  # supports 1, 2, 4 length

            bbox[..., 0:4:2] += padding_left
            bbox[..., 1:4:2] += padding_top

        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "(padding={0}, fill={1}, padding_mode={2})".format(
            self.padding, self.fill, self.padding_mode)


class Resize(T.Resize):
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
        interpolation: str,
        max_size: Optional[int],
    ):
        interpolation = INVERSE_MODES_MAPPING[interpolation]
        # @illian01: antialias paramter always true (always use) for PIL image, and NetsPresso Trainer uses PIL image for augmentation.
        super().__init__(size, interpolation, max_size)

    def forward(self, image, mask=None, bbox=None):
        w, h = image.size

        image = F.resize(image, self.size, self.interpolation, self.max_size, self.antialias)
        if mask is not None:
            mask = F.resize(mask, self.size, interpolation=T.InterpolationMode.NEAREST,
                            max_size=self.max_size)
        if bbox is not None:
            target_w, target_h = image.size # @illian01: Determine ratio according to the actual resized image
            bbox[..., 0:4:2] *= float(target_w / w)
            bbox[..., 1:4:2] *= float(target_h / h)
        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "(size={0}, interpolation={1}, max_size={2}, antialias={3})".format(
            self.size, self.interpolation.value, self.max_size, self.antialias)


class RandomHorizontalFlip:
    visualize = True

    def __init__(
        self,
        p: float,
    ):
        self.p: float = max(0., min(1., p))

    def __call__(self, image, mask=None, bbox=None):
        w, _ = image.size
        if random.random() < self.p:
            image = F.hflip(image)
            if mask is not None:
                mask = F.hflip(mask)
            if bbox is not None:
                bbox[..., 2::-2] = w - bbox[..., 0:4:2]
        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "(p={0})".format(self.p)


class RandomVerticalFlip:
    visualize = True

    def __init__(
        self,
        p: float,
    ):
        self.p: float = max(0., min(1., p))

    def __call__(self, image, mask=None, bbox=None):
        _, h = image.size
        if random.random() < self.p:
            image = F.vflip(image)
            if mask is not None:
                mask = F.vflip(mask)
            if bbox is not None:
                bbox[..., 3::-2] = h - bbox[..., 1:4:2]
        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "(p={0})".format(self.p)


class PadIfNeeded:
    visualize = True

    def __init__(self, size, fill=0, padding_mode="constant"):
        super().__init__()
        if not isinstance(size, (int, Sequence)):
            raise TypeError("Size should be int or sequence. Got {}".format(type(size)))
        if isinstance(size, Sequence) and len(size) not in (1, 2):
            raise ValueError("If size is a sequence, it should have 1 or 2 values")
        self.new_h = size[0] if isinstance(size, Sequence) else size
        self.new_w = size[1] if isinstance(size, Sequence) else size
        self.fill = fill
        self.padding_mode = padding_mode

    def __call__(self, image, mask=None, bbox=None):
        if not isinstance(image, (torch.Tensor, Image.Image)):
            raise TypeError("Image should be Tensor or PIL.Image. Got {}".format(type(image)))

        if isinstance(image, Image.Image):
            w, h = image.size
        else:
            w, h = image.shape[-1], image.shape[-2]

        w_pad_needed = max(0, self.new_w - w)
        h_pad_needed = max(0, self.new_h - h)
        padding_ltrb = [w_pad_needed // 2,
                        h_pad_needed // 2,
                        w_pad_needed // 2 + w_pad_needed % 2,
                        h_pad_needed // 2 + h_pad_needed % 2]
        image = F.pad(image, padding_ltrb, fill=self.fill, padding_mode=self.padding_mode)
        if mask is not None:
            mask = F.pad(mask, padding_ltrb, fill=255, padding_mode=self.padding_mode)
        if bbox is not None:
            padding_left, padding_top, _, _ = padding_ltrb
            bbox[..., 0:4:2] += padding_left
            bbox[..., 1:4:2] += padding_top
        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "(size={0}, fill={1}, padding_mode={2})".format(
            (self.new_h, self.new_w), self.fill, self.padding_mode
        )


class ColorJitter(T.ColorJitter):
    visualize = True

    def __init__(
        self,
        brightness: Union[float, List],
        contrast: Union[float, List],
        saturation: Union[float, List],
        hue: Union[float, List],
        p: float
    ):
        super(ColorJitter, self).__init__(brightness, contrast, saturation, hue)
        self.p: float = max(0., min(1., p))

    def forward(self, image, mask=None, bbox=None):
        fn_idx, brightness_factor, contrast_factor, saturation_factor, hue_factor = \
            self.get_params(self.brightness, self.contrast, self.saturation, self.hue)

        if random.random() < self.p:
            for fn_id in fn_idx:
                if fn_id == 0 and brightness_factor is not None:
                    image = F.adjust_brightness(image, brightness_factor)
                elif fn_id == 1 and contrast_factor is not None:
                    image = F.adjust_contrast(image, contrast_factor)
                elif fn_id == 2 and saturation_factor is not None:
                    image = F.adjust_saturation(image, saturation_factor)
                elif fn_id == 3 and hue_factor is not None:
                    image = F.adjust_hue(image, hue_factor)

        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + \
            f"(brightness={self.brightness}, " + \
            f"contrast={self.contrast}, " + \
            f"saturation={self.saturation}, " + \
            f"hue={self.hue}, " + \
            f"p={self.p})"


class RandomCrop:
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
    ):

        if not isinstance(size, (int, Sequence)):
            raise TypeError("Size should be int or sequence. Got {}".format(type(size)))
        if isinstance(size, Sequence) and len(size) not in (1, 2):
            raise ValueError("If size is a sequence, it should have 1 or 2 values")
        self.size_h = size[0] if isinstance(size, Sequence) else size
        self.size_w = size[1] if isinstance(size, Sequence) else size
        self.image_pad_if_needed = PadIfNeeded((self.size_h, self.size_w))

    def _crop_bbox(self, bbox, i, j, h, w):
        area_original = (bbox[..., 2] - bbox[..., 0]) * (bbox[..., 3] - bbox[..., 1])

        bbox[..., 0:4:2] = np.clip(bbox[..., 0:4:2] - j, 0, w)
        bbox[..., 1:4:2] = np.clip(bbox[..., 1:4:2] - i, 0, h)

        area_cropped = (bbox[..., 2] - bbox[..., 0]) * (bbox[..., 3] - bbox[..., 1])
        area_ratio = area_cropped / (area_original + 1)  # +1 for preventing ZeroDivisionError

        bbox = bbox[area_ratio >= BBOX_CROP_KEEP_THRESHOLD, ...]
        return bbox

    def __call__(self, image, mask=None, bbox=None):
        image, mask, bbox = self.image_pad_if_needed(image=image, mask=mask, bbox=bbox)
        i, j, h, w = T.RandomCrop.get_params(image, (self.size_h, self.size_w))
        image = F.crop(image, i, j, h, w)
        if mask is not None:
            mask = F.crop(mask, i, j, h, w)
        if bbox is not None:
            bbox_candidate = self._crop_bbox(bbox, i, j, h, w)
            _bbox_crop_count = 1
            while bbox_candidate.shape[0] == 0:
                if _bbox_crop_count == MAX_RETRY:
                    raise ValueError(f"It seems no way to use crop augmentation for this dataset. bbox: {bbox}, (i, j, h, w): {(i, j, h, w)}")
                bbox_candidate = self._crop_bbox(bbox, i, j, h, w)
                _bbox_crop_count += 1
            bbox = bbox_candidate
        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "(size={0})".format((self.size_h, self.size_w))


class RandomResizedCrop(T.RandomResizedCrop):
    visualize = True

    def __init__(
        self,
        size: Union[int, List],
        scale: Union[float, List],
        ratio: Union[float, List],
        interpolation: str,
    ):
        interpolation = INVERSE_MODES_MAPPING[interpolation]
        # @illian01: antialias paramter always true (always use) for PIL image, and NetsPresso Trainer uses PIL image for augmentation.
        super().__init__(size, scale, ratio, interpolation)

    def _crop_bbox(self, bbox, i, j, h, w):
        area_original = (bbox[..., 2] - bbox[..., 0]) * (bbox[..., 3] - bbox[..., 1])

        bbox[..., 0:4:2] = np.clip(bbox[..., 0:4:2] - j, 0, w)
        bbox[..., 1:4:2] = np.clip(bbox[..., 1:4:2] - i, 0, h)

        area_cropped = (bbox[..., 2] - bbox[..., 0]) * (bbox[..., 3] - bbox[..., 1])
        area_ratio = area_cropped / (area_original + 1)  # +1 for preventing ZeroDivisionError

        bbox = bbox[area_ratio >= BBOX_CROP_KEEP_THRESHOLD, ...]
        return bbox

    def forward(self, image, mask=None, bbox=None):
        w_orig, h_orig = image.size
        i, j, h, w = self.get_params(image, self.scale, self.ratio)
        image = F.resized_crop(image, i, j, h, w, self.size, self.interpolation)
        if mask is not None:
            mask = F.resized_crop(mask, i, j, h, w, self.size, interpolation=T.InterpolationMode.NEAREST)
        if bbox is not None:
            # img = crop(img, top, left, height, width)
            bbox_candidate = self._crop_bbox(bbox, i, j, h, w)
            _bbox_crop_count = 1
            while bbox_candidate.shape[0] != 0:
                if _bbox_crop_count == MAX_RETRY:
                    raise ValueError(f"It seems no way to use crop augmentation for this dataset. bbox: {bbox}, (i, j, h, w): {(i, j, h, w)}")
                bbox_candidate = self._crop_bbox(bbox, i, j, h, w)
                _bbox_crop_count += 1
            bbox = bbox_candidate

            # img = resize(img, size, interpolation)
            w_cropped, h_cropped = np.clip(w_orig - j, 0, w), np.clip(h_orig - i, 0, h)
            target_w, target_h = (self.size, self.size) if isinstance(self.size, int) else self.size
            bbox[..., 0:4:2] *= float(target_w / w_cropped)
            bbox[..., 1:4:2] *= float(target_h / h_cropped)

        return image, mask, bbox

    def __repr__(self):
        interpolate_str = self.interpolation.value
        format_string = self.__class__.__name__ + '(size={0}'.format(self.size)
        format_string += ', scale={0}'.format(tuple(round(s, 4) for s in self.scale))
        format_string += ', ratio={0}'.format(tuple(round(r, 4) for r in self.ratio))
        format_string += ', interpolation={0})'.format(interpolate_str)
        return format_string


class RandomErasing(T.RandomErasing):
    visualize = True

    def __init__(
        self,
        p: float,
        scale: List,
        ratio: List,
        value: Optional[int],
        inplace: bool,
    ):
        scale = list(scale)
        ratio = list(ratio)
        super().__init__(p, scale, ratio, value, inplace)

    @staticmethod
    def get_params(
        img, scale: Tuple[float, float], ratio: Tuple[float, float], value: Optional[int] = None
    ):
        img_w, img_h = img.size

        area = img_h * img_w

        log_ratio = torch.log(torch.tensor(ratio))
        for _ in range(10):
            erase_area = area * torch.empty(1).uniform_(scale[0], scale[1]).item()
            aspect_ratio = torch.exp(torch.empty(1).uniform_(log_ratio[0], log_ratio[1])).item()

            h = int(round(math.sqrt(erase_area * aspect_ratio)))
            w = int(round(math.sqrt(erase_area / aspect_ratio)))
            if not (h < img_h and w < img_w):
                continue

            if value is None:
                v = np.random.randint(255, size=(h, w)).astype('uint8')
                v = Image.fromarray(v).convert(img.mode)
            else:
                v = Image.new(img.mode, (w, h), value)

            i = torch.randint(0, img_h - h + 1, size=(1,)).item()
            j = torch.randint(0, img_w - w + 1, size=(1,)).item()
            return i, j, v

        # Return original image
        return 0, 0, img

    def forward(self, image, mask=None, bbox=None):
        if torch.rand(1) < self.p:
            x, y, v = self.get_params(image, scale=self.scale, ratio=self.ratio, value=self.value)
            image.paste(v, (y, x))
            # TODO: Object-aware
            return image, mask, bbox
        return image, mask, bbox


class TrivialAugmentWide(torch.nn.Module):
    """
    Based on the torchvision implementation.
    https://pytorch.org/vision/main/_modules/torchvision/transforms/autoaugment.html#TrivialAugmentWide
    """
    visualize = True

    def __init__(
        self,
        num_magnitude_bins: int,
        interpolation: str,
        fill: Optional[Union[List[int], int]],
    ) -> None:
        super().__init__()
        interpolation = INVERSE_MODES_MAPPING[interpolation]

        self.num_magnitude_bins = num_magnitude_bins
        self.interpolation = interpolation
        self.fill = fill

    def _augmentation_space(self, num_bins: int) -> Dict[str, Tuple[Tensor, bool]]:
        return {
            # op_name: (magnitudes, signed)
            "Identity": (torch.tensor(0.0), False),
            "ShearX": (torch.linspace(0.0, 0.99, num_bins), True),
            "ShearY": (torch.linspace(0.0, 0.99, num_bins), True),
            "TranslateX": (torch.linspace(0.0, 32.0, num_bins), True),
            "TranslateY": (torch.linspace(0.0, 32.0, num_bins), True),
            "Rotate": (torch.linspace(0.0, 135.0, num_bins), True),
            "Brightness": (torch.linspace(0.0, 0.99, num_bins), True),
            "Color": (torch.linspace(0.0, 0.99, num_bins), True),
            "Contrast": (torch.linspace(0.0, 0.99, num_bins), True),
            "Sharpness": (torch.linspace(0.0, 0.99, num_bins), True),
            "Posterize": (8 - (torch.arange(num_bins) / ((num_bins - 1) / 6)).round().int(), False),
            "Solarize": (torch.linspace(255.0, 0.0, num_bins), False),
            "AutoContrast": (torch.tensor(0.0), False),
            "Equalize": (torch.tensor(0.0), False),
        }

    def forward(self, image, mask=None, bbox=None):
        fill = self.fill
        channels, height, width = F.get_dimensions(image)
        if isinstance(image, Tensor):
            if isinstance(fill, (int, float)):
                fill = [float(fill)] * channels
            elif fill is not None:
                fill = [float(f) for f in fill]

        op_meta = self._augmentation_space(self.num_magnitude_bins)
        op_index = int(torch.randint(len(op_meta), (1,)).item())
        op_name = list(op_meta.keys())[op_index]
        magnitudes, signed = op_meta[op_name]
        magnitude = (
            float(magnitudes[torch.randint(len(magnitudes), (1,), dtype=torch.long)].item())
            if magnitudes.ndim > 0
            else 0.0
        )
        if signed and torch.randint(2, (1,)):
            magnitude *= -1.0

        # TODO: Compute mask, bbox
        return _apply_op(image, op_name, magnitude, interpolation=self.interpolation, fill=fill), mask, bbox

    def __repr__(self) -> str:
        s = (
            f"{self.__class__.__name__}("
            f"num_magnitude_bins={self.num_magnitude_bins}"
            f", interpolation={self.interpolation}"
            f", fill={self.fill}"
            f")"
        )
        return s


class AutoAugment(T.AutoAugment):
    visualize = True

    def __init__(
        self,
        policy: str,
        interpolation: str,
        fill: Optional[List[float]],
    ) -> None:
        policy = T.AutoAugmentPolicy(policy)
        interpolation = INVERSE_MODES_MAPPING[interpolation]

        super().__init__(policy, interpolation, fill)

    def forward(self, image, mask=None, bbox=None):
        """
            img (PIL Image or Tensor): Image to be transformed.

        Returns:
            PIL Image or Tensor: AutoAugmented image.
        """
        fill = self.fill
        channels, height, width = F.get_dimensions(image)
        if isinstance(image, Tensor):
            if isinstance(fill, (int, float)):
                fill = [float(fill)] * channels
            elif fill is not None:
                fill = [float(f) for f in fill]

        transform_id, probs, signs = self.get_params(len(self.policies))

        op_meta = self._augmentation_space(10, (height, width))
        for i, (op_name, p, magnitude_id) in enumerate(self.policies[transform_id]):
            if probs[i] <= p:
                magnitudes, signed = op_meta[op_name]
                magnitude = float(magnitudes[magnitude_id].item()) if magnitude_id is not None else 0.0
                if signed and signs[i] == 0:
                    magnitude *= -1.0
                image = _apply_op(image, op_name, magnitude, interpolation=self.interpolation, fill=fill)

        # TODO: Compute mask, bbox
        return image, mask, bbox


class RandomMixup:
    """
    Based on the RandomMixup implementation of ml_cvnets.
    https://github.com/apple/ml-cvnets/blob/77717569ab4a852614dae01f010b32b820cb33bb/data/transforms/image_torch.py

    Given a batch of input images and labels, this class randomly applies the
    `MixUp transformation <https://arxiv.org/abs/1710.09412>`_

    Args:
        opts (argparse.Namespace): Arguments
        num_classes (int): Number of classes in the dataset
    """
    visualize = False

    def __init__(
        self,
        num_classes: int,
        alpha: float,
        p: float,
        inplace: bool,
    ):
        if not (num_classes > 0):
            raise ValueError("Please provide a valid positive value for the num_classes.")
        if not (alpha > 0):
            raise ValueError("Alpha param can't be zero.")
        if not (0.0 < p <= 1.0):
            raise ValueError("MixUp probability should be between 0 and 1, where 1 is inclusive")

        self.num_classes = num_classes
        self.alpha = alpha
        self.p = p
        self.inplace = inplace

    def _apply_mixup_transform(self, image_tensor, target_tensor):
        if image_tensor.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {image_tensor.ndim}")
        if target_tensor.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target_tensor.ndim}")
        if not image_tensor.is_floating_point():
            raise ValueError(f"Batch datatype should be a float tensor. Got {image_tensor.dtype}.")
        if target_tensor.dtype != torch.int64:
            raise ValueError(f"Target datatype should be torch.int64. Got {target_tensor.dtype}")

        if not self.inplace:
            image_tensor = image_tensor.clone()
            target_tensor = target_tensor.clone()

        if target_tensor.ndim == 1:
            target_tensor = F_torch.one_hot(
                target_tensor, num_classes=self.num_classes
            ).to(dtype=image_tensor.dtype)

        # It's faster to roll the batch by one instead of shuffling it to create image pairs
        batch_rolled = image_tensor.roll(1, 0)
        target_rolled = target_tensor.roll(1, 0)

        # Implemented as on mixup paper, page 3.
        lambda_param = float(
            torch._sample_dirichlet(torch.tensor([self.alpha, self.alpha]))[0]
        )
        batch_rolled.mul_(1.0 - lambda_param)
        image_tensor.mul_(lambda_param).add_(batch_rolled)

        target_rolled.mul_(1.0 - lambda_param)
        target_tensor.mul_(lambda_param).add_(target_rolled)
        return image_tensor, target_tensor

    def __call__(self, samples, targets):
        if torch.rand(1).item() >= self.p:
            return samples, targets

        mixup_samples, mixup_targets = self._apply_mixup_transform(
            image_tensor=samples, target_tensor=targets
        )

        return mixup_samples, mixup_targets

    def __repr__(self) -> str:
        return "{}(num_classes={}, p={}, alpha={}, inplace={})".format(
            self.__class__.__name__, self.num_classes, self.p, self.alpha, self.inplace
        )


class RandomCutmix:
    """
    Based on the RandomCutmix implementation of ml_cvnets.
    https://github.com/apple/ml-cvnets/blob/77717569ab4a852614dae01f010b32b820cb33bb/data/transforms/image_torch.py

    Given a batch of input images and labels, this class randomly applies the
    `CutMix transformation <https://arxiv.org/abs/1905.04899>`_

    Args:
        opts (argparse.Namespace): Arguments
        num_classes (int): Number of classes in the dataset
    """
    visualize = False

    def __init__(
        self,
        num_classes: int,
        alpha: float,
        p: float,
        inplace: bool,
    ):
        if not (num_classes > 0):
            raise ValueError("Please provide a valid positive value for the num_classes.")
        if not (alpha > 0):
            raise ValueError("Alpha param can't be zero.")
        if not (0.0 < p <= 1.0):
            raise ValueError("CutMix probability should be between 0 and 1, where 1 is inclusive")

        self.num_classes = num_classes
        self.alpha = alpha
        self.p = p
        self.inplace = inplace

    def _apply_cutmix_transform(self, image_tensor, target_tensor):
        if image_tensor.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {image_tensor.ndim}")
        if target_tensor.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target_tensor.ndim}")
        if not image_tensor.is_floating_point():
            raise ValueError(f"Batch dtype should be a float tensor. Got {image_tensor.dtype}.")
        if target_tensor.dtype != torch.int64:
            raise ValueError(f"Target dtype should be torch.int64. Got {target_tensor.dtype}")

        if not self.inplace:
            image_tensor = image_tensor.clone()
            target_tensor = target_tensor.clone()

        if target_tensor.ndim == 1:
            target_tensor = F_torch.one_hot(
                target_tensor, num_classes=self.num_classes
            ).to(dtype=image_tensor.dtype)

        # It's faster to roll the batch by one instead of shuffling it to create image pairs
        batch_rolled = image_tensor.roll(1, 0)
        target_rolled = target_tensor.roll(1, 0)

        # Implemented as on cutmix paper, page 12 (with minor corrections on typos).
        lambda_param = float(
            torch._sample_dirichlet(torch.tensor([self.alpha, self.alpha]))[0]
        )
        W, H = F.get_image_size(image_tensor)

        r_x = torch.randint(W, (1,))
        r_y = torch.randint(H, (1,))

        r = 0.5 * math.sqrt(1.0 - lambda_param)
        r_w_half = int(r * W)
        r_h_half = int(r * H)

        x1 = int(torch.clamp(r_x - r_w_half, min=0))
        y1 = int(torch.clamp(r_y - r_h_half, min=0))
        x2 = int(torch.clamp(r_x + r_w_half, max=W))
        y2 = int(torch.clamp(r_y + r_h_half, max=H))

        image_tensor[:, :, y1:y2, x1:x2] = batch_rolled[:, :, y1:y2, x1:x2]
        lambda_param = float(1.0 - (x2 - x1) * (y2 - y1) / (W * H))

        target_rolled.mul_(1.0 - lambda_param)
        target_tensor.mul_(lambda_param).add_(target_rolled)
        return image_tensor, target_tensor

    def __call__(self, samples, targets) -> Dict:
        if torch.rand(1).item() >= self.p:
            return samples, targets

        mixup_samples, mixup_targets = self._apply_cutmix_transform(
            image_tensor=samples, target_tensor=targets
        )

        return mixup_samples, mixup_targets

    def __repr__(self) -> str:
        return "{}(num_classes={}, p={}, alpha={}, inplace={})".format(
            self.__class__.__name__, self.num_classes, self.p, self.alpha, self.inplace
        )


class Normalize:
    visualize = False

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, image, mask=None, bbox=None):
        image = F.normalize(image, mean=self.mean, std=self.std)
        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "(mean={0}, std={1})".format(
            self.mean, self.std
        )


class ToTensor(T.ToTensor):
    visualize = False

    def __call__(self, image, mask=None, bbox=None):
        image = F.to_tensor(image)
        if mask is not None:
            mask = torch.as_tensor(np.array(mask), dtype=torch.int64)
        if bbox is not None:
            bbox = torch.as_tensor(np.array(bbox), dtype=torch.float)

        return image, mask, bbox

    def __repr__(self):
        return self.__class__.__name__ + "()"



import os
from pathlib import Path
from typing import List, Literal, Optional, Union

import torch
from omegaconf import DictConfig, OmegaConf

from netspresso_trainer.cfg import TrainerConfig
from netspresso_trainer.trainer_util import (
    get_gpus_from_parser_and_config,
    parse_gpu_ids,
    train_with_config_impl,
    train_with_yaml_impl,
)

LOG_LEVEL = os.getenv('LOG_LEVEL', 'INFO')


def export_config_as_yaml(config: TrainerConfig) -> str:
    conf: DictConfig = OmegaConf.create(config)
    return OmegaConf.to_yaml(conf)


def train_with_config(
    config: TrainerConfig,
    gpus: Optional[str] = None,
    log_level: Literal['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'] = 'INFO'
) -> None:

    gpus: Union[List, int] = parse_gpu_ids(gpus)

    logging_dir: Path = train_with_config_impl(
        gpus=gpus,
        config=config,
        log_level=log_level
    )

    return logging_dir


def train_with_yaml(
    data: Union[Path, str],
    augmentation: Union[Path, str],
    model: Union[Path, str], training: Union[Path, str],
    logging: Union[Path, str], environment: Union[Path, str],
    gpus: Optional[str] = None, log_level: str = LOG_LEVEL
):

    gpus: Union[List, int] = parse_gpu_ids(gpus)

    logging_dir: Path = train_with_yaml_impl(
        gpus=gpus,
        data=data,
        augmentation=augmentation,
        model=model,
        training=training,
        logging=logging,
        environment=environment,
        log_level=log_level
    )

    return logging_dir



from .dataset import ClassficationDataSampler, classification_mix_collate_fn, classification_onehot_collate_fn
from .huggingface import ClassificationHFDataset
from .local import ClassificationCustomDataset



from typing import Literal

import numpy as np
import PIL.Image as Image

from ..augmentation.transforms import generate_edge, reduce_label
from ..base import BaseHFDataset


class SegmentationHFDataset(BaseHFDataset):

    def __init__(
            self,
            conf_data,
            conf_augmentation,
            model_name,
            idx_to_class,
            split,
            huggingface_dataset,
            transform=None,
            with_label=True,
            **kwargs
    ):
        root = conf_data.metadata.repo
        super(SegmentationHFDataset, self).__init__(
            conf_data,
            conf_augmentation,
            model_name,
            root,
            split,
            with_label
        )

        self.transform = transform
        self.idx_to_class = idx_to_class
        self.samples = huggingface_dataset

        assert "label_value_to_idx" in kwargs
        self.label_value_to_idx = kwargs["label_value_to_idx"]

        self.label_image_mode: Literal['RGB', 'L', 'P'] = str(conf_data.label_image_mode).upper() \
            if conf_data.label_image_mode is not None else 'L'

        self.image_feature_name = conf_data.metadata.features.image
        self.label_feature_name = conf_data.metadata.features.label

    @property
    def num_classes(self):
        return len(self.idx_to_class)

    @property
    def class_map(self):
        return self.idx_to_class

    def __len__(self):
        return self.samples.num_rows

    def __getitem__(self, index):

        img_name = f"{index:06d}"
        img: Image.Image = self.samples[index][self.image_feature_name]
        label: Image.Image = self.samples[index][self.label_feature_name] if self.label_feature_name in self.samples[index] else None

        label_array = np.array(label.convert(self.label_image_mode))
        label_array = label_array[..., np.newaxis] if label_array.ndim == 2 else label_array
        # if self.conf_augmentation.reduce_zero_label:
        #     label = reduce_label(np.array(label))

        mask = np.zeros((label.size[1], label.size[0]), dtype=np.uint8)
        for label_value in self.label_value_to_idx:
            class_mask = (label_array == np.array(label_value)).all(axis=-1)
            mask[class_mask] = self.label_value_to_idx[label_value]
        mask = Image.fromarray(mask, mode='L')  # single mode array (PIL.Image) compatbile with torchvision transform API

        org_img = img.copy()
        w, h = img.size

        if label is None:
            out = self.transform(self.conf_augmentation)(image=img)
            return {'pixel_values': out['image'], 'name': img_name, 'org_img': org_img, 'org_shape': (h, w)}

        outputs = {}

        if self.model_name == 'pidnet':
            edge = generate_edge(np.array(label))
            out = self.transform(self.conf_augmentation)(image=img, mask=mask, edge=edge)
            outputs.update({'pixel_values': out['image'], 'labels': out['mask'], 'edges': out['edge'].float(), 'name': img_name})
        else:
            out = self.transform(self.conf_augmentation)(image=img, mask=mask)
            outputs.update({'pixel_values': out['image'], 'labels': out['mask'], 'name': img_name})

        if self._split in ['train', 'training']:
            return outputs

        assert self._split in ['val', 'valid', 'test']
        # outputs.update({'org_img': org_img, 'org_shape': (h, w)})  # TODO: return org_img with batch_size > 1
        outputs.update({'org_shape': (h, w)})
        return outputs



from .base import BaseCSVLogger


class ClassificationCSVLogger(BaseCSVLogger):
    csv_header = ['epoch', 'train_loss', 'valid_loss', 'train_accuracy', 'valid_accuracy']
    def __init__(self, model, result_dir):
        super(ClassificationCSVLogger, self).__init__(model, result_dir)
        self.update_header(self.csv_header)

        self._key_map = {
            'epoch': 'epoch',
            'train/total': 'train_loss',
            'valid/total': 'valid_loss',
            'train/Acc@1': 'train_accuracy',
            'valid/Acc@1': 'valid_accuracy',
        }

    @property
    def key_map(self):
        return self._key_map

class SegmentationCSVLogger(BaseCSVLogger):
    csv_header = ['epoch', 'train_loss', 'valid_loss', 'train_miou', 'valid_miou', 'train_pixel_accuracy', 'valid_pixel_accuracy']
    def __init__(self, model, result_dir):
        super(SegmentationCSVLogger, self).__init__(model, result_dir)
        self.update_header(self.csv_header)

        self._key_map = {
            'epoch': 'epoch',
            'train/total': 'train_loss',
            'valid/total': 'valid_loss',
            'train/iou': 'train_miou',
            'valid/iou': 'valid_miou',
            'train/pixel_acc': 'train_pixel_accuracy',
            'valid/pixel_acc': 'valid_pixel_accuracy',
        }

    @property
    def key_map(self):
        return self._key_map


class DetectionCSVLogger(BaseCSVLogger):
    csv_header = ['epoch', 'train_loss', 'valid_loss', 'train_map50', 'valid_map50', 'train_map75', 'valid_map75', 'train_map50_95', 'valid_map50_95']
    def __init__(self, model, result_dir):
        super(DetectionCSVLogger, self).__init__(model, result_dir)
        self.update_header(self.csv_header)

        self._key_map = {
            'epoch': 'epoch',
            'train/total': 'train_loss',
            'valid/total': 'valid_loss',
            'train/map50': 'train_map50',
            'valid/map50': 'valid_map50',
            'train/map75': 'train_map75',
            'valid/map75': 'valid_map75',
            'train/map50_95': 'train_map50_95',
            'valid/map50_95': 'valid_map50_95',
        }

    @property
    def key_map(self):
        return self._key_map



from .registry import SUPPORTING_TASK_LIST, TASK_PIPELINE


def build_pipeline(conf, task, model_name, model, devices,
                   train_dataloader, eval_dataloader, class_map, logging_dir,
                   is_graphmodule_training, profile=False):
    assert task in SUPPORTING_TASK_LIST, f"No such task! (task: {task})"

    task_pipeline = TASK_PIPELINE[task]

    trainer = task_pipeline(conf, task, model_name, model, devices,
                            train_dataloader, eval_dataloader, class_map, logging_dir,
                            is_graphmodule_training=is_graphmodule_training, profile=profile)

    return trainer



import csv
import random
from collections import Counter
from itertools import chain
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import torch
from loguru import logger
from omegaconf import DictConfig
from torch.nn import functional as F
from torch.utils.data import random_split

from ..base import BaseDataSampler
from ..utils.constants import IMG_EXTENSIONS
from ..utils.misc import natural_key

VALID_IMG_EXTENSIONS = IMG_EXTENSIONS + tuple((x.upper() for x in IMG_EXTENSIONS))

def load_class_map_with_id_mapping(root_dir, train_dir,
                                   map_or_filename: Optional[Union[str, Path]]=None,
                                   id_mapping: Optional[Dict[str, str]]=None):

    if map_or_filename is None:  # may be labeled with directory
        # dir ->
        dir_list = [x.name for x in Path(train_dir).iterdir() if x.is_dir()]
        dir_to_class = id_mapping if id_mapping is not None else {k: k for k in dir_list}  # id_mapping or identity

        class_list = [dir_to_class[dir] for dir in dir_list]
        class_list = sorted(class_list, key=lambda k: natural_key(k))
        _class_to_idx = {class_name: class_idx for class_idx, class_name in enumerate(class_list)}
        idx_to_class = {v: k for k, v in _class_to_idx.items()}

        file_or_dir_to_idx = {dir: _class_to_idx[dir_to_class[dir]] for dir in dir_list}  # dir -> idx
        return file_or_dir_to_idx, idx_to_class

    # Assume the `map_or_filename` is path for csv label file
    class_map_path = Path(root_dir) / map_or_filename
    assert class_map_path.exists(), f"Cannot locate specified class map file {class_map_path}!"

    class_map_ext = class_map_path.suffix.lower()
    assert class_map_ext == '.csv', f"Unsupported class map file extension ({class_map_ext})!"

    with open(class_map_path, newline='') as csvfile:
        reader = csv.DictReader(csvfile)
        file_class_list = [{column: str(row[column]).strip() for column in ['image_id', 'class']}
                           for row in reader]

    class_stats = Counter([x['class'] for x in file_class_list])

    _class_to_idx = {class_name: class_idx
                    for class_idx, class_name in enumerate(sorted(class_stats, key=lambda k: natural_key(k)))}
    idx_to_class = {v: k for k, v in _class_to_idx.items()}

    file_or_dir_to_idx = {elem['image_id']: _class_to_idx[elem['class']] for elem in file_class_list}  # file -> idx

    return file_or_dir_to_idx, idx_to_class

def is_file_dict(image_dir: Union[Path, str], file_or_dir_to_idx):
    image_dir = Path(image_dir)
    candidate_name = list(file_or_dir_to_idx.keys())[0]
    file_or_dir: Path = image_dir / candidate_name
    if file_or_dir.exists():
        return file_or_dir.is_file()

    file_candidates = list(image_dir.glob(f"{candidate_name}.*"))
    assert len(file_candidates) != 0, f"Unknown label format! Is there any something file like {file_or_dir} ?"

    return True


def classification_mix_collate_fn(original_batch, mix_transforms):
    images = []
    target = []
    for data_sample in original_batch:
        images.append(data_sample[0])
        target.append(data_sample[1])

    images = torch.stack(images, dim=0)
    target = torch.tensor(target, dtype=torch.long)

    _mix_transform = random.choice(mix_transforms)
    images, target = _mix_transform(images, target)

    outputs = (images, target)
    return outputs


def classification_onehot_collate_fn(original_batch, num_classes):
    images = []
    target = []
    for data_sample in original_batch:
        images.append(data_sample[0])
        target.append(data_sample[1])

    images = torch.stack(images, dim=0)
    target = torch.tensor(target, dtype=torch.long)
    target = F.one_hot(target, num_classes=num_classes).to(dtype=images.dtype)

    outputs = (images, target)
    return outputs


class ClassficationDataSampler(BaseDataSampler):
    def __init__(self, conf_data, train_valid_split_ratio):
        super(ClassficationDataSampler, self).__init__(conf_data, train_valid_split_ratio)

    def load_data(self, file_or_dir_to_idx, split='train'):
        data_root = Path(self.conf_data.path.root)
        split_dir = self.conf_data.path[split]
        image_dir: Path = data_root / split_dir.image

        images_and_targets: List[Dict[str, Optional[Union[str, int]]]] = []

        assert split in ['train', 'valid', 'test'], f"split should be either {['train', 'valid', 'test']}"
        if split in ['train', 'valid']:

            if is_file_dict(image_dir, file_or_dir_to_idx):
                file_to_idx = file_or_dir_to_idx
                for ext in IMG_EXTENSIONS:
                    for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}')):
                        if file.name in file_to_idx:
                            images_and_targets.append({'image': str(file), 'label': file_to_idx[file.name]})
                            continue
                        logger.debug(f"Found file wihtout label: {file}")

            else:
                dir_to_idx = file_or_dir_to_idx
                for dir_name, dir_idx in dir_to_idx.items():
                    _dir = Path(image_dir) / dir_name
                    for ext in VALID_IMG_EXTENSIONS:
                        images_and_targets.extend([{'image': str(file), 'label': dir_idx} for file in chain(_dir.glob(f'*{ext}'), _dir.glob(f'*{ext.upper()}'))])

        else:  # split == test
            for ext in VALID_IMG_EXTENSIONS:
                images_and_targets.extend([{'image': str(file), 'label': None} for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}'))])


        images_and_targets = sorted(images_and_targets, key=lambda k: natural_key(k['image']))
        return images_and_targets

    def load_samples(self):
        assert self.conf_data.path.train.image is not None
        root_dir = Path(self.conf_data.path.root)
        train_dir = root_dir / self.conf_data.path.train.image
        id_mapping: Optional[dict] = dict(self.conf_data.id_mapping) if self.conf_data.id_mapping is not None else None
        file_or_dir_to_idx, idx_to_class = load_class_map_with_id_mapping(root_dir, train_dir, map_or_filename=self.conf_data.path.train.label, id_mapping=id_mapping)

        exists_valid = self.conf_data.path.valid.image is not None
        exists_test = self.conf_data.path.test.image is not None

        valid_samples = None
        test_samples = None

        train_samples = self.load_data(file_or_dir_to_idx, split='train')
        if exists_valid:
            valid_dir = root_dir / self.conf_data.path.valid.image
            file_or_dir_to_idx_valid, _ = load_class_map_with_id_mapping(root_dir, valid_dir, map_or_filename=self.conf_data.path.valid.label, id_mapping=id_mapping)
            valid_samples = self.load_data(file_or_dir_to_idx_valid, split='valid')
        if exists_test:
            test_samples = self.load_data(file_or_dir_to_idx, split='test')

        if not exists_valid:
            num_train_splitted = int(len(train_samples) * self.train_valid_split_ratio)
            train_samples, valid_samples = \
                random_split(train_samples, [num_train_splitted, len(train_samples) - num_train_splitted],
                                generator=torch.Generator().manual_seed(42))

        return train_samples, valid_samples, test_samples, {'idx_to_class': idx_to_class}

    def load_huggingface_samples(self):
        from datasets import ClassLabel, load_dataset

        cache_dir = self.conf_data.metadata.custom_cache_dir
        root = self.conf_data.metadata.repo
        subset_name = self.conf_data.metadata.subset
        if cache_dir is not None:
            cache_dir = Path(cache_dir)
            Path(cache_dir).mkdir(exist_ok=True, parents=True)
        total_dataset = load_dataset(root, name=subset_name, cache_dir=cache_dir)

        label_feature_name = self.conf_data.metadata.features.label
        label_feature = total_dataset['train'].features[label_feature_name]
        if isinstance(label_feature, ClassLabel):
            labels: List[str] = label_feature.names
        else:
            labels = list({sample[label_feature_name] for sample in total_dataset['train']})

        if isinstance(labels[0], int):
            # TODO: find class_map <-> idx and apply it (ex. using id_mapping)
            idx_to_class: Dict[int, int] = {k: k for k in labels}
        elif isinstance(labels[0], str):
            idx_to_class: Dict[int, str] = dict(enumerate(labels))

        exists_valid = 'validation' in total_dataset
        exists_test = 'test' in total_dataset

        train_samples = total_dataset['train']
        valid_samples = None
        if exists_valid:
            valid_samples = total_dataset['validation']
        test_samples = None
        if exists_test:
            test_samples = total_dataset['test']

        if not exists_valid:
            splitted_datasets = train_samples.train_test_split(test_size=(1 - self.train_valid_split_ratio))
            train_samples = splitted_datasets['train']
            valid_samples = splitted_datasets['test']
        return train_samples, valid_samples, test_samples, {'idx_to_class': idx_to_class}
